import math
from timeit import default_timer as timer
from typing import Any, Callable, Dict, Sequence

import pandas as pd

from woodwork.logical_types import (
    Age,
    AgeNullable,
    Datetime,
    Integer,
    IntegerNullable,
    Unknown,
)
from woodwork.statistics_utils._get_histogram_values import _get_histogram_values
from woodwork.statistics_utils._get_mode import _get_mode
from woodwork.statistics_utils._get_numeric_value_counts_in_range import (
    _get_numeric_value_counts_in_range,
)
from woodwork.statistics_utils._get_recent_value_counts import _get_recent_value_counts
from woodwork.statistics_utils._get_top_values_categorical import (
    _get_top_values_categorical,
)
from woodwork.utils import CallbackCaller, _is_latlong_nan


def percentile(N, percent, count):
    """
    Source: https://stackoverflow.com/questions/2374640/how-do-i-calculate-percentiles-with-python-numpy/2753343#2753343

    Find the percentile of a list of values.

    Args:
        N (pd.Series): Series is a list of values. Note N MUST BE already sorted.
        percent (float): float value from 0.0 to 1.0.
        count (int): Count of values in series

    Returns:
        Value at percentile.
    """
    k = (count - 1) * percent
    f = math.floor(k)
    c = math.ceil(k)
    if f == c:
        return N[int(k)]
    d0 = N.iat[int(f)] * (c - k)
    d1 = N.iat[int(c)] * (k - f)
    return d0 + d1


def _get_describe_dict(
    dataframe: pd.DataFrame,
    include: Sequence[str] = None,
    callback: Callable[[int, int, int, str, float], Any] = None,
    results_callback: Callable[[pd.DataFrame, pd.Series], Any] = None,
    extra_stats: bool = False,
    bins: int = 10,
    top_x: int = 10,
    recent_x: int = 10,
) -> Dict[str, dict]:
    """Calculates statistics for data contained in a DataFrame using Woodwork typing information.

    Args:
        dataframe (pd.DataFrame): DataFrame to be described with Woodwork typing information initialized
        include (list[str or LogicalType], optional): filter for what columns to include in the
            statistics returned. Can be a list of column names, semantic tags, logical types, or a list
            combining any of the three. It follows the most broad specification. Favors logical types
            then semantic tag then column name. If no matching columns are found, an empty DataFrame
            will be returned.
        callback (callable, optional): function to be called with incremental updates. Has the following parameters:

            - update (int): change in progress since last call
            - progress (int): the progress so far in the calculations
            - total (int): the total number of calculations to do
            - unit (str): unit of measurement for progress/total
            - time_elapsed (float): total time in seconds elapsed since start of call
        results_callback (callable, optional): function to be called with intermediate results. Has the following parameters:

            - results_so_far (pd.DataFrame): the full dataframe calculated so far
            - most_recent_calculation (pd.Series): the calculations for the most recent column

        extra_stats (bool): If True, will calculate a histogram for numeric columns, top values
            for categorical columns and value counts for the most recent values in datetime columns. Will also
            calculate value counts within the range of values present for integer columns if the range of
            values present is less than or equal to than the number of bins used to compute the histogram.
            Output can be controlled by bins, top_x and recent_x parameters.
        bins (int): Number of bins to use when calculating histogram for numeric columns. Defaults to 10.
            Will be ignored unless extra_stats=True.
        top_x (int): Number of items to return when getting the most frequently occurring values for categorical
            columns. Defaults to 10. Will be ignored unless extra_stats=True.
        recent_x (int): Number of values to return when calculating value counts for the most recent dates in
            datetime columns. Defaults to 10. Will be ignored unless extra_stats=True.

    Returns:
        dict[str -> dict]: A dictionary with a key for each column in the data or for each column
        matching the logical types, semantic tags or column names specified in ``include``, paired
        with a value containing a dictionary containing relevant statistics for that column.
    """
    start_time = timer()
    unit = "calculations"
    if include is not None:
        filtered_cols = dataframe.ww._filter_cols(include, col_names=True)
        cols_to_include = [
            (k, v) for k, v in dataframe.ww.columns.items() if k in filtered_cols
        ]
    else:
        cols_to_include = dataframe.ww.columns.items()

    results = {}

    # Setup for progress callback and make initial call
    # Assume 1 unit for general preprocessing, plus main loop over column
    total_loops = 1 + len(cols_to_include)
    callback_caller = CallbackCaller(callback, unit, total_loops, start_time=start_time)
    callback_caller.update(1)

    for column_name, column in cols_to_include:
        if "index" in column.semantic_tags:
            callback_caller.update(1)
            continue
        values = {}
        logical_type = column.logical_type
        semantic_tags = column.semantic_tags
        series = dataframe[column_name]

        agg_stats_to_calculate = {
            "category": ["count", "nunique"],
            Datetime: ["count", "max", "min", "nunique", "mean"],
            Unknown: ["count", "nunique"],
        }

        # Calculate Aggregation Stats
        if column.is_categorical:
            agg_stats = agg_stats_to_calculate["category"]
        elif column.is_numeric:
            agg_stats = None
        elif column.is_datetime:
            agg_stats = agg_stats_to_calculate[Datetime]
        elif column.is_unknown:
            agg_stats = agg_stats_to_calculate[Unknown]
        else:
            agg_stats = ["count"]
        values = {}
        if agg_stats:
            values = series.agg(agg_stats).to_dict()

        mode = _get_mode(series)

        if column.is_latlong:
            nan_count = series.apply(_is_latlong_nan).sum()
            count = len(series) - nan_count

            values["nan_count"] = nan_count
            values["count"] = count
        else:
            values["nan_count"] = series.isna().sum()

        # Calculate other specific stats based on logical type or semantic tags
        if column.is_boolean:
            values["num_false"] = series.value_counts().get(False, 0)
            values["num_true"] = series.value_counts().get(True, 0)
        elif column.is_numeric:
            if values["nan_count"] == 0:
                agg_stats = ["count", "nunique", "mean", "std"]
                values.update(series.agg(agg_stats).to_dict())
                series = series.sort_values(
                    ignore_index=True,
                )
                values["max"] = series.iat[int(values["count"] - 1)]
                values["min"] = series.iat[0]
                values["first_quartile"] = percentile(
                    series,
                    0.25,
                    int(values["count"]),
                )
                values["second_quartile"] = percentile(
                    series,
                    0.5,
                    int(values["count"]),
                )
                values["third_quartile"] = percentile(
                    series,
                    0.75,
                    int(values["count"]),
                )
            else:
                agg_stats = ["count", "max", "min", "nunique", "mean", "std"]
                values.update(series.agg(agg_stats).to_dict())
                quant_values = series.quantile([0.25, 0.5, 0.75]).tolist()
                values["first_quartile"] = quant_values[0]
                values["second_quartile"] = quant_values[1]
                values["third_quartile"] = quant_values[2]

        values["mode"] = mode
        values["physical_type"] = series.dtype
        values["logical_type"] = logical_type
        values["semantic_tags"] = semantic_tags

        # Calculate extra detailed stats, if requested
        if extra_stats:
            if column.is_numeric:
                if pd.isnull(values["max"]) or pd.isnull(values["min"]):
                    values["histogram"] = []
                    values["top_values"] = []
                else:
                    values["histogram"] = _get_histogram_values(series, bins=bins)
                    if isinstance(
                        column.logical_type,
                        (Age, AgeNullable, Integer, IntegerNullable),
                    ):
                        _range = range(int(values["min"]), int(values["max"]) + 1)
                        # Calculate top numeric values if range of values present
                        # is less than or equal number of histogram bins
                        range_len = int(values["max"]) + 1 - int(values["min"])
                        if range_len <= bins and (series % 1 == 0).all():
                            values["top_values"] = _get_numeric_value_counts_in_range(
                                series,
                                _range,
                            )
            elif column.is_categorical:
                values["top_values"] = _get_top_values_categorical(series, top_x)
            elif column.is_datetime:
                values["recent_values"] = _get_recent_value_counts(series, recent_x)

        results[column_name] = values
        if results_callback is not None:
            results_so_far = pd.DataFrame.from_dict(results)
            most_recent_calculations = pd.Series(values, name=column_name)
            results_callback(results_so_far, most_recent_calculations)
        callback_caller.update(1)
    return results



import pandas as pd

from woodwork.statistics_utils.frequency_inference._constants import (
    OBSERVED_COLUMN_NAME,
)
from woodwork.statistics_utils.frequency_inference._get_ranges import _get_ranges
from woodwork.statistics_utils.frequency_inference._types import RangeObject


def _determine_duplicate_values(observed_ts):
    """Calculate Duplicate Values in time_series:

    Args:
        observed_ts (Series): The observed time_series.

    Returns:
        (list(RangeObject)): A list of RangeObject data objects. A RangeObject has the following properties:

        - dt: an ISO 8601 formatted string of the duplicate timestamp
        - idx: first index of duplicate timestamp. Index is relative to observed timeseries
        - range: the number of sequential elements that are duplicated
    """

    observed_df = (
        pd.DataFrame({OBSERVED_COLUMN_NAME: observed_ts}).reset_index(drop=True).diff()
    )

    observed_dupes = observed_df[observed_df[OBSERVED_COLUMN_NAME] == pd.Timedelta(0)]

    if len(observed_dupes) == 0:
        return []

    ranges = _get_ranges(observed_dupes.index)

    return [
        RangeObject(
            dt=observed_ts[start_idx].isoformat(),
            idx=int(start_idx),
            range=int(end_idx - start_idx + 1),
        )
        for start_idx, end_idx in ranges
    ]



import sys

import numpy as np
import pandas as pd
import pytest

from woodwork.logical_types import Categorical, CountryCode, Double, Integer, Unknown
from woodwork.type_sys.inference_functions import (
    categorical_func,
    double_func,
    integer_func,
)
from woodwork.type_sys.type_system import INFERENCE_SAMPLE_SIZE, TypeSystem


# Integer Inference Fixtures
@pytest.fixture
def integers():
    return [
        pd.Series(4 * [-1, 2, 1, 7]),
        pd.Series(4 * [-1, 0, 5, 3]),
        pd.Series(4 * [sys.maxsize, -sys.maxsize - 1, 0], dtype="str").astype("int64"),
    ]


# Double Inference Fixtures
@pytest.fixture
def doubles():
    return [
        pd.Series(4 * [-1, 2.5, 1, 7]),
        pd.Series(4 * [1.5, np.nan, 1, 3]),
        pd.Series(4 * [1.5, np.inf, 1, 3]),
        pd.Series(4 * [np.finfo("d").max, np.finfo("d").min, 3, 1]),
    ]


# Boolean Inference Fixtures
@pytest.fixture
def bools():
    return [
        pd.Series([True, False, True, True]),
        pd.Series([True, np.nan, True, True]),
        pd.Series(["y", "n", "N", "Y"]),
        pd.Series(["True", "false", "FALSE", "TRUE"]),
        pd.Series(["t", "f", "T", "T"]),
        pd.Series(["yes", "no", "NO", "Yes"]),
        pd.Series(["y", "n", "N", None]),
        pd.Series(["True", "false", pd.NA, "TRUE"]),
        pd.Series(["t", "f", "T", np.nan]),
        pd.Series(["yes", "no", "NO", pd.NA]),
    ]


# Datetime Inference Fixtures
@pytest.fixture
def datetimes():
    return [
        pd.Series(["2000-3-11", "2000-3-12", "2000-03-13", "2000-03-14"]),
        pd.Series(["2000-3-11", np.nan, "2000-03-13", "2000-03-14"]),
    ]


# Email Inference Fixtures
@pytest.fixture
def emails():
    return [
        pd.Series(
            ["fl@alteryx.com", "good@email.com", "boaty@mcboatface.com", "foo@bar.com"],
        ),
        pd.Series(["fl@alteryx.com", "good@email.com", "boaty@mcboatface.com", np.nan]),
        pd.Series(
            ["fl@alteryx.com", "good@email.com"] * 2,
        ),
    ]


# Email Inference Fixtures
@pytest.fixture
def bad_emails():
    return [
        pd.Series(["fl@alteryx.com", "not_an_email", "good@email.com", "foo@bar.com"]),
        pd.Series(["fl@alteryx.com", "b☃d@email.com", "good@email.com", np.nan]),
        pd.Series(["fl@alteryx.com", "@email.com", "good@email.com", "foo@bar.com"]),
        pd.Series(["fl@alteryx.com", "bad@email", "good@email.com", np.nan]),
        pd.Series([np.nan, np.nan, np.nan, np.nan]),
        pd.Series([1, 2, 3, 4]).astype("int"),
        pd.Series([{"key": "value"}]).astype("O"),
        pd.Series([(1, 2), (3, 4)]).astype("O"),
    ]


# Categorical Inference Fixtures
@pytest.fixture
def categories():
    return [
        pd.Series(10 * ["a", "b", "a", "b"]),
        pd.Series(10 * ["1", "2", "1", "2"]),
        pd.Series(10 * ["a", np.nan, "b", "b"]),
        pd.Series(10 * [1, 2, 1, 2]),
    ]


@pytest.fixture
def categories_dtype():
    return pd.DataFrame(
        {
            "cat": pd.Series(["a", "b", "c", "d"], dtype="category"),
            "non_cat": pd.Series(["a", "b", "c", "d"], dtype="string"),
        },
    )


# Timedelta Inference Fixtures
@pytest.fixture
def timedeltas():
    return [
        pd.Series(pd.to_timedelta(range(4), unit="s")),
        pd.Series([pd.to_timedelta(1, unit="s"), np.nan]),
    ]


# Natural Language Fixtures
@pytest.fixture
def natural_language():
    return [
        pd.Series(
            [
                "Hello World! My name is bob!",
                "I like to move it move it",
                "its cold outside",
            ],
        ),
    ]


# Postal Inference Fixtures
@pytest.fixture
def postal():
    return [
        pd.Series(10 * ["77002", "55106"]),
        pd.Series(10 * ["77002-0000", "55106-0000"]),
        pd.Series(10 * ["12345", "12345", "12345-6789", "12345-0000"]),
    ]


# Unknown Inference Fixtures
@pytest.fixture
def strings():
    return [
        pd.Series(
            ["Mr. John Doe", "Doe, Mrs. Jane", "James Brown", "Ms. Paige Turner"],
        ),
        pd.Series(["y", "no", "N", None]),
        pd.Series(["T", "false", pd.NA, "TRUE"]),
        pd.Series(["1", "1", np.nan, "0.0"]),
    ]


# pd.NA Inference Fixtures
@pytest.fixture
def pdnas():
    return [
        pd.Series(
            [
                "Hello World! My name is bob!",
                pd.NA,
                "I like to move it move it",
                "its cold outside",
            ],
        ),
        pd.Series(["Mr. John Doe", pd.NA, "James Brown", "Ms. Paige Turner"]).astype(
            "string",
        ),
        pd.Series([1, pd.NA, 2, 3]).astype("Int64"),
        pd.Series([True, pd.NA, False, True]).astype("boolean"),
    ]


# Empty Series Inference Fixtures
@pytest.fixture
def empty_series():
    return pd.Series([], dtype="object")


# Null Inference Fixtures
@pytest.fixture
def nulls():
    return [
        pd.Series([pd.NA, pd.NA, pd.NA, pd.NA]),
        pd.Series([np.nan, np.nan, np.nan, np.nan]),
        pd.Series([None, None, None, None]),
        pd.Series([None, np.nan, pd.NA, None]),
        pd.Series(["None", "null", "n/a", "NAN"]),
    ]


@pytest.fixture
def large_df():
    df = pd.DataFrame()
    df["int_nullable"] = [int(i) for i in range(INFERENCE_SAMPLE_SIZE)] + [np.nan]
    df["bool_nullable"] = [True, False] * int(INFERENCE_SAMPLE_SIZE // 2) + [pd.NA]
    df["floats"] = [int(i) for i in range(INFERENCE_SAMPLE_SIZE)] + [1.2]
    df["constant"] = [None] * (INFERENCE_SAMPLE_SIZE + 1)
    return df


@pytest.fixture
def default_inference_functions():
    return {
        Double: double_func,
        Integer: integer_func,
        Categorical: categorical_func,
        CountryCode: None,
        Unknown: None,
    }


@pytest.fixture
def default_relationships():
    return [(Double, Integer), (Categorical, CountryCode)]


@pytest.fixture
def type_sys(default_inference_functions, default_relationships):
    return TypeSystem(
        inference_functions=default_inference_functions,
        relationships=default_relationships,
        default_type=Unknown,
    )


# URL Inference Fixtures
@pytest.fixture
def urls():
    return [
        pd.Series([f"http://url{i}.com" for i in range(100)]),
        pd.Series(
            [
                "http://url.com",
                "http://url.org",
                "https://another.net",
                "https://schoo.edu",
            ]
            * 25,
        ),
    ]


# Phone Number Inference Fixtures
@pytest.fixture
def phone():
    return [
        pd.Series([f"200.200.786{i}" for i in range(9)]),
        pd.Series(["311-311-3156", "(755) 755 7109", "+1(288)-288-7772"] * 3),
    ]


# IP Address Inference Fixtures
@pytest.fixture
def ip():
    return [
        pd.Series(
            [
                "172.16.254.1",
                "2001:0db8:85a3:0000:0000:8a2e:0370:7334",
                "1762:0:0:0:0:B03:1:AF18",
            ]
            * 2,
        ),
        pd.Series([f"172.16.254.{i}" for i in range(6)]),
    ]



import pandas as pd
from pandas.api import types as pdtypes

import woodwork as ww
from woodwork.logical_types import Categorical, Double, LogicalType


def test_register_custom_logical_type(type_sys):
    class CustomLogicalType(LogicalType):
        primary_dtype = "string"

    def custom_func(series):
        return True

    type_sys.add_type(
        CustomLogicalType,
        inference_function=custom_func,
        parent="Categorical",
    )
    assert CustomLogicalType in type_sys.registered_types
    assert (Categorical, CustomLogicalType) in type_sys.relationships
    assert type_sys.inference_functions[CustomLogicalType] is custom_func
    assert isinstance(
        type_sys.infer_logical_type(pd.Series(["a", "b"] + 10 * ["a"])),
        CustomLogicalType,
    )


def test_custom_type_with_accessor(sample_df):
    class AgesAbove20(LogicalType):
        primary_dtype = "float64"
        standard_tags = {"age", "numeric"}

    def ages_func(series):
        if all(series.dropna() > 20):
            return True
        return False

    ww.type_system.add_type(
        AgesAbove20,
        inference_function=ages_func,
        parent="IntegerNullable",
    )
    sample_df.ww.init()
    assert isinstance(sample_df.ww["age"].ww.logical_type, AgesAbove20)
    assert sample_df.ww["age"].ww.semantic_tags == {"age", "numeric"}
    assert sample_df["age"].dtype == "float64"
    # Reset global type system to original settings
    ww.type_system.reset_defaults()


def test_accessor_override_default_function(sample_df):
    def new_double_func(series):
        if pdtypes.is_bool_dtype(series.dtype):
            return True
        return False

    # Update functions to cause 'is_registered' to be recognized as Double instead fo BooleanNullable
    ww.type_system.update_inference_function(
        "Double",
        inference_function=new_double_func,
    )
    ww.type_system.update_inference_function("BooleanNullable", inference_function=None)
    sample_df.ww.init()
    assert isinstance(sample_df.ww["is_registered"].ww.logical_type, Double)
    assert sample_df["is_registered"].dtype == "float64"
    # Reset global type system to original settings
    ww.type_system.reset_defaults()



from timeit import default_timer as timer
from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest
from packaging import version

import woodwork as ww
from woodwork.exceptions import TypeValidationError
from woodwork.logical_types import (
    Age,
    AgeFractional,
    AgeNullable,
    Boolean,
    BooleanNullable,
    Categorical,
    CountryCode,
    CurrencyCode,
    Datetime,
    Double,
    Integer,
    IntegerNullable,
    Ordinal,
    PostalCode,
    SubRegionCode,
)
from woodwork.type_sys.type_system import DEFAULT_INFERENCE_FUNCTIONS
from woodwork.type_sys.utils import (
    _get_specified_ltype_params,
    _is_categorical_series,
    _is_numeric_series,
    col_is_datetime,
    list_logical_types,
    list_semantic_tags,
)
from woodwork.utils import (
    CallbackCaller,
    _coerce_to_float,
    _convert_input_to_set,
    _get_column_logical_type,
    _infer_datetime_format,
    _is_latlong_nan,
    _is_nan,
    _is_s3,
    _is_url,
    _is_valid_latlong_series,
    _is_valid_latlong_value,
    _parse_logical_type,
    _reformat_to_latlong,
    camel_to_snake,
    get_valid_mi_types,
    import_or_none,
    import_or_raise,
)


def test_camel_to_snake():
    test_items = {
        "PostalCode": "postal_code",
        "SubRegionCode": "sub_region_code",
        "NaturalLanguage": "natural_language",
        "Categorical": "categorical",
    }

    for key, value in test_items.items():
        assert camel_to_snake(key) == value


def test_convert_input_to_set():
    error_message = "semantic_tags must be a string, set or list"
    with pytest.raises(TypeError, match=error_message):
        _convert_input_to_set(int)

    error_message = "test_text must be a string, set or list"
    with pytest.raises(TypeError, match=error_message):
        _convert_input_to_set({"index": {}, "time_index": {}}, "test_text")

    error_message = "include parameter must contain only strings"
    with pytest.raises(TypeError, match=error_message):
        _convert_input_to_set(["index", 1], "include parameter")

    semantic_tags_from_single = _convert_input_to_set("index", "include parameter")
    assert semantic_tags_from_single == {"index"}

    semantic_tags_from_list = _convert_input_to_set(["index", "numeric", "category"])
    assert semantic_tags_from_list == {"index", "numeric", "category"}

    semantic_tags_from_set = _convert_input_to_set(
        {"index", "numeric", "category"},
        "include parameter",
    )
    assert semantic_tags_from_set == {"index", "numeric", "category"}


@patch("woodwork.utils._validate_string_tags")
@patch("woodwork.utils._validate_tags_input_type")
def test_validation_methods_called(mock_validate_input_type, mock_validate_strings):
    assert not mock_validate_input_type.called
    assert not mock_validate_strings.called

    _convert_input_to_set("test_tag", validate=False)
    assert not mock_validate_input_type.called

    _convert_input_to_set("test_tag", validate=True)
    assert mock_validate_input_type.called

    _convert_input_to_set(["test_tag", "tag2"], validate=False)
    assert not mock_validate_strings.called

    _convert_input_to_set(["test_tag", "tag2"], validate=True)
    assert mock_validate_strings.called


def test_list_logical_types_default():
    all_ltypes = ww.logical_types.LogicalType.__subclasses__()

    df = list_logical_types()

    assert set(df.columns) == {
        "name",
        "type_string",
        "description",
        "physical_type",
        "standard_tags",
        "is_default_type",
        "is_registered",
        "parent_type",
    }

    assert len(all_ltypes) == len(df)
    default_types_set = {str(cls) for cls in DEFAULT_INFERENCE_FUNCTIONS.keys()}
    listed_as_default = set(df[df["is_default_type"]]["name"])
    assert listed_as_default == default_types_set


def test_list_logical_types_customized_type_system():
    ww.type_system.remove_type("URL")

    class CustomRegistered(ww.logical_types.LogicalType):
        primary_dtype = "int64"

    class CustomNotRegistered(ww.logical_types.LogicalType):
        primary_dtype = "int64"

    ww.type_system.add_type(CustomRegistered)
    all_ltypes = ww.logical_types.LogicalType.__subclasses__()
    df = list_logical_types()
    assert len(all_ltypes) == len(df)
    # Check that URL is unregistered
    url = df[df.name == "URL"].iloc[0]
    assert url.is_default_type
    assert not url.is_registered

    # Check that new registered type is present and shows as registered
    index = df.name == "CustomRegistered"
    assert index.any()
    custom = df[index].iloc[0]
    assert not custom.is_default_type
    assert custom.is_registered

    # Check that new unregistered type is present and shows as not registered
    index = df.name == "CustomNotRegistered"
    assert index.any()
    custom = df[index].iloc[0]
    assert not custom.is_default_type
    assert not custom.is_registered
    ww.type_system.reset_defaults()


def test_list_semantic_tags():
    df = list_semantic_tags()

    assert set(df.columns) == {"name", "is_standard_tag", "valid_logical_types"}

    for name, valid_ltypes in df[["name", "valid_logical_types"]].values:
        if name in ["passthrough", "ignore", "index"]:
            assert valid_ltypes == "Any LogicalType"
        elif name not in ["time_index", "date_of_birth"]:
            assert isinstance(valid_ltypes, list)
            for log_type in valid_ltypes:
                assert name in log_type.standard_tags


def test_is_numeric_datetime_series(time_index_df):
    assert _is_numeric_series(time_index_df["ints"], None)
    assert _is_numeric_series(time_index_df["ints"], Double)
    assert not _is_numeric_series(time_index_df["ints"], Categorical)
    assert _is_numeric_series(time_index_df["ints"], Datetime)

    assert _is_numeric_series(time_index_df["strs"], Integer)
    assert not _is_numeric_series(time_index_df["strs"], "Categorical")
    assert not _is_numeric_series(time_index_df["strs"], Categorical)
    assert _is_numeric_series(time_index_df["strs"], Double)
    assert _is_numeric_series(time_index_df["strs"], "Double")

    assert not _is_numeric_series(time_index_df["bools"], None)
    assert not _is_numeric_series(time_index_df["bools"], "Boolean")

    assert not _is_numeric_series(time_index_df["times"], None)
    assert not _is_numeric_series(time_index_df["times"], Datetime)

    assert not _is_numeric_series(time_index_df["letters"], None)
    assert not _is_numeric_series(time_index_df["letters"], Double)
    assert not _is_numeric_series(time_index_df["letters"], Categorical)


def test_get_ltype_params():
    params_empty_class = _get_specified_ltype_params(Categorical)
    assert params_empty_class == {}
    params_empty = _get_specified_ltype_params(Categorical())
    assert params_empty == {}

    params_class = _get_specified_ltype_params(Datetime)
    assert params_class == {}

    params_null = _get_specified_ltype_params(Datetime())
    assert params_null == {"datetime_format": None, "timezone": None}

    ymd = "%Y-%m-%d"
    params_value = _get_specified_ltype_params(
        Datetime(datetime_format=ymd, timezone="UTC"),
    )
    assert params_value == {"datetime_format": ymd, "timezone": "UTC"}


def test_import_or_raise():
    assert import_or_raise("pandas", "Module pandas could not be found") == pd

    error = "Module nonexistent could not be found."
    with pytest.raises(ImportError, match=error):
        import_or_raise("nonexistent", error)


def test_import_or_none():
    assert import_or_none("pandas") == pd
    assert import_or_none("nonexistent") is None


def test_is_url():
    assert _is_url("https://www.google.com/")
    assert not _is_url("google.com")


def test_is_s3():
    assert _is_s3("s3://test-bucket/test-key")
    assert not _is_s3("https://woodwork-static.s3.amazonaws.com/")


@pytest.mark.parametrize(
    "test_input, error_msg",
    [
        ({1, 2, 3}, "LatLong value is not properly formatted."),
        ("{1, 2, 3}", "LatLong value is not properly formatted."),
        ("This is text", "LatLong value is not properly formatted."),
        ("'(1,2)'", "LatLong value is not properly formatted."),
        ((1, 2, 3), "LatLong values must have exactly two values"),
        ("(1, 2, 3)", "LatLong values must have exactly two values"),
        ("(1,)", "LatLong values must have exactly two values"),
        (
            ("41deg52'54\" N", "21deg22'54\" W"),
            "LatLong values must be in decimal degrees.",
        ),
        ((41.5, "21deg22'54\" W"), "LatLong values must be in decimal degrees."),
    ],
)
def test_reformat_to_latlong_errors(test_input, error_msg):
    with pytest.raises(TypeValidationError, match=error_msg):
        _reformat_to_latlong(test_input)


@pytest.mark.parametrize(
    "test_input,expected",
    [
        ((1, 2), (1, 2)),
        (("1", "2"), (1, 2)),
        ("(1,2)", (1, 2)),
        ([1, 2], (1, 2)),
        (["1", "2"], (1, 2)),
        ("[1, 2]", (1, 2)),
        ("1, 2", (1, 2)),
        ((1, np.nan), (1, np.nan)),
        ((np.nan, "1"), (np.nan, 1)),
        ("(1, nan)", (1, np.nan)),
        ("(NaN, 9)", (np.nan, 9)),
        ("(1, None)", (1, np.nan)),
        ("(<NA>, 9)", (np.nan, 9)),
        ((np.nan, np.nan), (np.nan, np.nan)),
        ((pd.NA, pd.NA), (np.nan, np.nan)),
        ((None, None), (np.nan, np.nan)),
        (None, np.nan),
        (np.nan, np.nan),
        (pd.NA, np.nan),
        ("None", np.nan),
        ("NaN", np.nan),
        ("<NA>", np.nan),
    ],
)
def test_reformat_to_latlong(test_input, expected):
    if isinstance(expected, (list, tuple)):
        assert _reformat_to_latlong(test_input) == expected
    else:
        assert _reformat_to_latlong(test_input) is expected


@pytest.mark.parametrize(
    "test_input,expected",
    [
        (4, 4.0),
        ("2.2", 2.2),
        (None, np.nan),
        (np.nan, np.nan),
        (pd.NA, np.nan),
    ],
)
def test_coerce_to_float(test_input, expected):
    if np.isnan(expected):
        assert _coerce_to_float(test_input) is expected
    else:
        assert _coerce_to_float(test_input) == expected


def test_is_nan():
    assert _is_nan(None)
    assert _is_nan(np.nan)
    assert _is_nan(pd.NA)
    assert _is_nan("None")
    assert _is_nan("nan")
    assert _is_nan("NaN")
    assert _is_nan("<NA>")
    assert _is_nan("")
    assert _is_nan("null")

    assert not _is_nan([None, 1, 3])
    assert not _is_nan([])
    assert not _is_nan(0)
    assert not _is_nan(False)
    assert not _is_nan({"key": "value"})


@pytest.mark.parametrize(
    "test_input,expected",
    [
        ((1.0, 2.0), True),
        ((1.0, np.nan), True),
        ((np.nan, 2.0), True),
        ((np.nan, np.nan), True),
        (np.nan, True),
        (pd.NA, False),
        (2.0, False),
        ([2.0], False),
        ([None, None], False),
        ("None", False),
        ([1.0, 2.0], False),
        ((pd.NA, pd.NA), True),
        (("a", 2.0), False),
        ((1.0, 2.0, 3.0), False),
        (None, False),
    ],
)
def test_is_valid_latlong_value(test_input, expected):
    assert _is_valid_latlong_value(test_input) == expected


def test_is_valid_latlong_series():
    valid_series = pd.Series([(1.0, 2.0), (3.0, 4.0)])
    invalid_series = pd.Series([(1.0, 2.0), (3.0, "4.0")])

    assert _is_valid_latlong_series(valid_series) is True
    assert _is_valid_latlong_series(invalid_series) is False


def test_get_valid_mi_types():
    valid_types = get_valid_mi_types()
    expected_types = [
        Age,
        AgeFractional,
        AgeNullable,
        Boolean,
        BooleanNullable,
        Categorical,
        CountryCode,
        CurrencyCode,
        Datetime,
        Double,
        Integer,
        IntegerNullable,
        Ordinal,
        PostalCode,
        SubRegionCode,
    ]

    assert valid_types == expected_types


def test_get_column_logical_type(sample_series):
    assert isinstance(
        _get_column_logical_type(sample_series, None, "col_name"),
        Categorical,
    )

    assert isinstance(
        _get_column_logical_type(sample_series, Datetime, "col_name"),
        Datetime,
    )


def test_parse_logical_type():
    assert isinstance(_parse_logical_type("Datetime", "col_name"), Datetime)
    assert isinstance(_parse_logical_type(Datetime, "col_name"), Datetime)

    ymd_format = Datetime(datetime_format="%Y-%m-%d")
    assert _parse_logical_type(ymd_format, "col_name") == ymd_format


def test_parse_logical_type_errors():
    error = "Invalid logical type specified for 'col_name'"
    with pytest.raises(TypeError, match=error):
        _parse_logical_type(int, "col_name")


def test_col_is_datetime():
    inputs = [
        pd.to_datetime(pd.Series(["2020-01-01", "2021-02-02", "2022-03-03"])),
        pd.to_datetime(pd.Series([pd.NA, "2021-02-02", "2022-03-03"])),
        pd.Series([1, 2, 3]),
        pd.Series([pd.NA, 2, 3]),
        pd.Series([1.0, 2.0, 3.0]),
        pd.Series([pd.NA, 2.0, 3.0]),
        pd.Series(["2020-01-01", "2021-02-02", "2022-03-03"]),
        pd.Series([pd.NA, "2021-02-02", "2022-03-03"]),
        pd.Series(["a", "b", "c"]),
        pd.Series([pd.NA, "b", "c"]),
        pd.Series([pd.NA, pd.NA, pd.NA]),
        pd.Series([1950.5, 1951, 1953.2, 1955.3]),
        pd.Series([1950, 1951, 1953, 1955]),
        pd.Series([1950.5, 1951, 1953.2, np.nan]),
        pd.Series([1950, 1951, 1953, np.nan]),
        pd.Series([1950.5, 1951, 1953.2, pd.NA]),
        pd.Series([1950, 1951, 1953, pd.NA]),
        pd.Series([1950, 1951, 1953, None]),
    ]

    expected_values = [
        True,
        True,
        False,
        False,
        False,
        False,
        True,
        True,
        False,
        False,
        False,
        False,
        False,
        False,
        False,
        False,
        False,
        False,
    ]

    for input, expected in list(zip(inputs, expected_values)):
        actual = col_is_datetime(input)
        assert actual is expected


def test_infer_datetime_format(datetimes):
    for series in datetimes:
        fmt = _infer_datetime_format(series)
        assert fmt == "%m/%d/%Y"

    dt = pd.Series(
        ["3/11/2000 9:00", "3/11/2000 10:00", "3/11/2000 11:00", "3/11/2000 12:00"],
    )
    fmt = _infer_datetime_format(dt)
    assert fmt == "%m/%d/%Y %H:%M"

    if version.parse(pd.__version__) >= version.parse("2.0.2"):
        # There was a pandas bug in earlier versions that was casing the
        # hour to get inferred incorrectly as %H (0-24) when it should have
        # been inferred as %I (0-12) when AM/PM is present.
        # Pandas Issue: https://github.com/pandas-dev/pandas/issues/53147
        # https://github.com/alteryx/woodwork/pull/1158
        dt = pd.Series(["Tue 24 Aug 2021 01:30:48 AM"])
        fmt = _infer_datetime_format(dt)
        assert fmt == "%a %d %b %Y %I:%M:%S %p"

        # https://github.com/alteryx/woodwork/pull/1158
        dt = pd.Series(["Tuesday 24 Aug 2021 01:30:48 AM"])
        fmt = _infer_datetime_format(dt)
        assert fmt == "%A %d %b %Y %I:%M:%S %p"

    dt = pd.Series(["Tue 24 Aug 2021 01:30:48"])
    fmt = _infer_datetime_format(dt)
    assert fmt == "%a %d %b %Y %H:%M:%S"

    dt = pd.Series(
        ["13-03-16 10:30:15+0000", "14-03-16 10:30:15+0000", "15-03-16 10:30:15+0000"],
    )
    fmt = _infer_datetime_format(dt)
    assert fmt == "%y-%m-%d %H:%M:%S%z"

    dt = pd.Series(
        [
            pd.Timestamp("13-03-16 10:30:15+0000"),
            pd.Timestamp("14-03-16 10:30:15+0000"),
            pd.Timestamp("15-03-16 10:30:15+0000"),
        ],
    )
    fmt = _infer_datetime_format(dt)
    assert fmt == "%Y-%m-%d %H:%M:%S%z"


def test_infer_datetime_format_all_null():
    missing_data = [
        pd.Series([None, None, None]),
        pd.Series([np.nan, np.nan, np.nan]),
        pd.Series([pd.NA, pd.NA, pd.NA]),
        pd.Series([]),
    ]

    for pd_series in missing_data:
        assert _infer_datetime_format(pd_series) is None


def test_is_categorical() -> None:
    # not categorical because unhashable type (list)
    assert not _is_categorical_series(pd.Series([[1]]), 0)
    assert not _is_categorical_series(pd.Series([None, [1]]), 0)

    # not categorical because empty series
    assert not _is_categorical_series(pd.Series([]), 0)
    assert not _is_categorical_series(pd.Series([None]), 0)
    assert not _is_categorical_series(pd.Series([None, None]), 0)

    # not categorical because too many unique values
    assert not _is_categorical_series(pd.Series([1, 2]), 0.5)
    assert not _is_categorical_series(pd.Series([1, 2, 3, 1]), 0.5)
    assert not _is_categorical_series(pd.Series([1, 2, 3, 4]), 0.75)

    # categorical
    assert _is_categorical_series(pd.Series([1, 1]), 0.5)
    assert _is_categorical_series(pd.Series([1, 2, 1, 1]), 0.5)
    assert _is_categorical_series(pd.Series([1, 2, 3, 1]), 0.75)


def test_is_latlong_nan():
    assert _is_latlong_nan((np.nan, np.nan))
    assert _is_latlong_nan([np.nan, np.nan])
    assert _is_latlong_nan((np.nan,))
    assert _is_latlong_nan(np.nan)
    assert not _is_latlong_nan((np.nan, 2.0))
    assert not _is_latlong_nan([np.nan, 2.0])
    assert not _is_latlong_nan((2.0, 3.0))
    assert not _is_latlong_nan("test")


def test_callback_caller(mock_callback):
    unit = "yards"
    total = 100
    caller = CallbackCaller(callback=mock_callback, unit=unit, total=total)
    start_time = caller.start_time

    assert mock_callback.total_update == 0
    assert caller.total == total
    assert caller.unit == unit
    assert caller.current_progress == 0
    assert isinstance(start_time, float)

    caller.update(50)
    assert mock_callback.unit == unit
    assert mock_callback.total_update == 50
    assert caller.current_progress == 50
    assert mock_callback.total_elapsed_time > 0
    assert mock_callback.progress_history == [50]


def test_callback_caller_defaults(mock_callback):
    unit = "yards"
    total = 100
    start_time = timer()
    caller = CallbackCaller(
        callback=mock_callback,
        unit=unit,
        total=total,
        start_time=start_time,
        start_progress=20,
    )

    assert caller.start_time == start_time
    assert caller.current_progress == 20


def test_callback_caller_no_callback():
    unit = "yards"
    total = 100
    caller = CallbackCaller(callback=None, unit=unit, total=total)

    assert caller.callback is None
    assert caller.total == 100
    assert caller.current_progress == 0

    caller.update(1)

    assert caller.current_progress == 0



import pandas as pd

from woodwork.deserializers.deserializer_base import Deserializer


class CSVDeserializer(Deserializer):
    """Deserialize Woodwork table from serialized data in CSV format."""

    format = "csv"

    def read_from_local_path(self):
        return pd.read_csv(self.read_path, dtype=self.column_dtypes, **self.kwargs)



import pandas as pd

from woodwork.deserializers.deserializer_base import Deserializer


class PickleDeserializer(Deserializer):
    """Deserialize Woodwork table from serialized data in pickle format."""

    format = "pickle"

    def read_from_local_path(self):
        return pd.read_pickle(self.read_path, **self.kwargs)



import pandas as pd

from woodwork.deserializers.deserializer_base import Deserializer


class FeatherDeserializer(Deserializer):
    """Deserialize Woodwork table from serialized data in feather format."""

    format = "feather"

    def read_from_local_path(self):
        return pd.read_feather(self.read_path)



from collections import namedtuple

import numpy as np
import pandas as pd
from scipy.stats import skew

import woodwork as ww
from woodwork.statistics_utils._get_medcouple_statistic import _sample_for_medcouple

method_result = namedtuple(
    "MedcoupleHeuristicResult",
    (
        "method",
        "mc",
    ),
)


def _determine_best_outlier_method(series):
    """Determines the best outlier method to use based on the distribution of the series and outcome of the medcouple statistic.

    Args:
        series (Series): Data on which the medcouple statistic will be run in order to determine skewness.

    Note:
        The calculation of the medcouple statistic has a large memory requirement of O(N**2), therefore larger series will
        have a random subset selected in order to determine skewness and the best outlier method.

    Returns:
        MedcoupleHeuristicResult - Named tuple with 2 fields
            method (str): Name of the outlier method to use.
            mc (float): The medcouple statistic (if the method chosen is medcouple, otherwise None).
    """
    mc = _sample_for_medcouple(series)
    method = "medcouple"
    if np.abs(mc) < ww.config.get_option("medcouple_threshold"):
        method = "box_plot"
        mc = None
    return method_result(method, mc)


def _determine_coefficients(series, mc):
    """Determines the coefficients to use in the medcouple bounds equation using the moment of skewness.

    Args:
        series (Series): Data on which the medcouple statistic will be run in order to determine coefficients.

    Note:
        Based off of the paper I. H. Adil and A. R. Irshad, "A Modified Approach for Detection of Outliers",
        Pak.j.stat.oper.res. Vol.XI No.1 2015 pp. 91-102, 2015.

    Returns:
        MedcoupleHeuristicResult - Named tuple with 2 fields
            method (str): Name of the outlier method to use.
            mc (float): The medcouple statistic (if the method chosen is medcouple, otherwise None).
    """
    try:
        coeff = np.abs(skew(series))
    except ValueError:
        # skew can't handle Int64 dtype
        coeff = np.abs(skew(series.astype("float64")))
    coeff = min(coeff, 3.5)
    if mc >= 0:
        return -coeff, coeff
    else:
        return coeff, -coeff


def _get_low_high_bound(series, method, q1, q3, min_value, max_value, mc=None):
    iqr = q3 - q1
    if method == "medcouple":
        if mc is None:
            raise ValueError(
                "If the method selected is medcouple, then mc cannot be None.",
            )
        # Medcouple coefficients change based on the skew direction
        lower_bound_coef, higher_bound_coef = _determine_coefficients(series, mc)
        low_bound = q1 - 1.5 * np.exp(lower_bound_coef * mc) * iqr
        high_bound = q3 + 1.5 * np.exp(higher_bound_coef * mc) * iqr
    elif method == "box_plot":
        low_bound = q1 - (iqr * 1.5)
        high_bound = q3 + (iqr * 1.5)
    else:
        raise ValueError(
            f"Acceptable methods are 'box_plot' and 'medcouple'. The value passed was '{method}'.",
        )
    # Box plot bounds calculation - the bounds should never be beyond the min and max values
    low_bound = max(low_bound, min_value)
    high_bound = min(high_bound, max_value)
    return low_bound, high_bound


def _get_box_plot_info_for_column(
    series,
    method="best",
    quantiles=None,
    include_indices_and_values=True,
    ignore_zeros=False,
):
    """Gets the information necessary to create a box and whisker plot with outliers for a numeric column.

    Args:
        series (Series): Data for which the box plot and outlier information will be gathered.
            Will be used to calculate quantiles if none are provided.
        method (str): The method to use when calculating the box and whiskers plot. Options are 'box_plot' and 'medcouple'.
        Defaults to 'best' at which point a heuristic will determine the appropriate method to use.
        quantiles (dict[float -> float], optional): A dictionary containing the quantiles for the data
            where the key indicates the quantile, and the value is the quantile's value for the data.
        include_indices_and_values (bool, optional): Whether or not the lists containing individual
            outlier values and their indices will be included in the returned dictionary.
            Defaults to True.
        ignore_zeros (bool): Whether to ignore 0 values (not NaN values) when calculating the box plot and outliers.
                Defaults to False.

    Note:
        The minimum quantiles necessary for building a box plot using the IQR method are the
        minimum value (0.0 in the quantiles dict), first quartile (0.25), third quartile (0.75), and maximum value (1.0).
        If no quantiles are provided, the following quantiles will be calculated:
        {0.0, 0.25, 0.5, 0.75, 1.0}, which correspond to {min, first quantile, median, third quantile, max}.

    Returns:
        (dict[str -> float,list[number]]): Returns a dictionary containing box plot information for the Series.
            The following elements will be found in the dictionary:

            - low_bound (float): the lowest data point in the dataset excluding any outliers - to be used as a whisker
            - high_bound (float): the highest point in the dataset excluding any outliers - to be used as a whisker
            - quantiles (list[float]): the quantiles used to determine the bounds.
                If quantiles were passed in, will contain all quantiles passed in. Otherwise, contains the five
                quantiles {0.0, 0.25, 0.5, 0.75, 1.0}.
            - low_values (list[float, int], optional): the values of the lower outliers.
                Will not be included if ``include_indices_and_values`` is False.
            - high_values (list[float, int], optional): the values of the upper outliers
                Will not be included if ``include_indices_and_values`` is False.
            - low_indices (list[int], optional): the corresponding index values for each of the lower outliers
                Will not be included if ``include_indices_and_values`` is False.
            - high_indices (list[int], optional): the corresponding index values for each of the upper outliers
                Will not be included if ``include_indices_and_values`` is False.
    """
    if not series.ww._schema.is_numeric:
        raise TypeError("Cannot calculate box plot statistics for non-numeric column")

    if quantiles and not isinstance(quantiles, dict):
        raise TypeError("quantiles must be a dictionary.")

    # remove null values from the data
    series = series.dropna()

    if ignore_zeros:
        series = series[series.astype(bool)]

    outliers_dict = {}
    # An empty or fully null Series has no outliers, bounds, or quantiles
    if series.shape[0] == 0:
        if include_indices_and_values:
            outliers_dict = {
                "low_values": [],
                "high_values": [],
                "low_indices": [],
                "high_indices": [],
            }
        return {
            "low_bound": np.nan,
            "high_bound": np.nan,
            "quantiles": {
                0.0: np.nan,
                0.25: np.nan,
                0.5: np.nan,
                0.75: np.nan,
                1.0: np.nan,
            },
            **outliers_dict,
        }

    # calculate the outlier bounds using IQR
    if quantiles is None:
        quantiles = series.quantile([0.0, 0.25, 0.5, 0.75, 1.0]).to_dict()
    elif len(set(quantiles.keys()) & {0.0, 0.25, 0.75, 1.0}) != 4:
        raise ValueError(
            "Input quantiles do not contain the minimum necessary quantiles for box plot calculation: "
            "0.0 (the minimum value), 0.25 (the first quartile), 0.75 (the third quartile), and 1.0 (the maximum value).",
        )
    min_value = quantiles[0.0]
    q1 = quantiles[0.25]
    q3 = quantiles[0.75]
    max_value = quantiles[1.0]

    mc = None
    if method == "best":
        method, mc = _determine_best_outlier_method(series)
    if method == "medcouple" and mc is None:
        mc = _sample_for_medcouple(series)

    if isinstance(mc, np.ndarray):
        mc = mc.tolist()

    low_bound, high_bound = _get_low_high_bound(
        series,
        method,
        q1,
        q3,
        min_value,
        max_value,
        mc,
    )
    low_bound = round(low_bound, 5) if isinstance(low_bound, float) else low_bound
    high_bound = round(high_bound, 5) if isinstance(high_bound, float) else high_bound

    if include_indices_and_values:
        # identify outliers in the series
        low_series = (
            series[series < low_bound]
            if low_bound > min_value
            else pd.Series(dtype="float64")
        )
        high_series = (
            series[series > high_bound]
            if high_bound < max_value
            else pd.Series(dtype="float64")
        )

        outliers_dict = {
            "low_values": low_series.tolist(),
            "high_values": high_series.tolist(),
            "low_indices": low_series.index.tolist(),
            "high_indices": high_series.index.tolist(),
        }

    outliers_dict["method"] = method
    if method == "medcouple":
        outliers_dict["medcouple_stat"] = mc

    return {
        "low_bound": low_bound,
        "high_bound": high_bound,
        "quantiles": quantiles,
        **outliers_dict,
    }



import contextlib

import numpy as np
import pandas as pd

CONFIG_DEFAULTS = {
    "categorical_threshold": 0.2,
    "numeric_categorical_threshold": None,
    "email_inference_regex": r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)",
    "url_inference_regex": r"(http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+)",
    "ipv4_inference_regex": r"(^(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])$)",
    "ipv6_inference_regex": "".join(
        r"""
(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]
{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|
([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]
{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}
|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9])
{0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}
[0-9]){0,1}[0-9]))
""".splitlines(),
    ),
    "phone_inference_regex": r"(?:\+?(0{2})?1[-.\s●]?)?\(?([2-9][0-9]{2})\)?[-\.\s●]?([2-9][0-9]{2})[-\.\s●]?([0-9]{4})$",
    "postal_code_inference_regex": r"^[0-9]{5}(?:-[0-9]{4})?$",
    "nan_values": [
        "",
        " ",
        None,
        np.nan,
        pd.NaT,
        "None",
        "NONE",
        "none",
        "NULL",
        "Null",
        "null",
        "NAN",
        "NaN",
        "Nan",
        "nan",
        "NA",
        "na",
        "N/A",
        "n/a",
        "n/A",
        "N/a",
        "<NA>",
        "<N/A>",
        "<n/a>",
        "<na>",
    ],
    "frequence_inference_window_length": 15,
    "frequence_inference_threshold": 0.9,
    "correlation_metrics": ["mutual_info", "pearson", "spearman", "max", "all"],
    "medcouple_threshold": 0.3,  # Must be between 0.0 and 1.0
    "medcouple_sample_size": 10000,
    "boolean_inference_strings": {
        frozenset(["yes", "no"]),
        frozenset(["y", "n"]),
        frozenset(["true", "false"]),
        frozenset(["t", "f"]),
    },
    "boolean_transform_mappings": {
        "yes": True,
        "no": False,
        "y": True,
        "n": False,
        "true": True,
        "false": False,
        "t": True,
        "f": False,
    },
    # when adding to boolean_inference_ints, add `0, 1` to the set directly
    "boolean_inference_ints": {},
}


class Config:
    def __init__(self, default_values):
        self._defaults = default_values
        self._data = default_values.copy()

    def set_option(self, key, value):
        if key not in self._data.keys():
            raise KeyError(f"Invalid option specified: {key}")
        self._data[key] = value

    def get_option(self, key):
        if key not in self._data.keys():
            raise KeyError(f"Invalid option specified: {key}")
        return self._data[key]

    def reset_option(self, key):
        if key not in self._data.keys():
            raise KeyError(f"Invalid option specified: {key}")
        self._data[key] = self._defaults[key]

    @contextlib.contextmanager
    def with_options(self, **options):
        old_options = {k: self.get_option(k) for k in options}

        for k, v in options.items():
            self.set_option(k, v)
        try:
            yield
        finally:
            for k, v in old_options.items():
                self.set_option(k, v)

    def __repr__(self):
        output_string = "Woodwork Global Config Settings\n"
        output_string += "-" * (len(output_string) - 1)
        for key, value in self._data.items():
            output_string += f"\n{key}: {value}"
        return output_string


config = Config(CONFIG_DEFAULTS)



import re
from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest

from woodwork.accessor_utils import (
    _is_dataframe,
    init_series,
)
from woodwork.column_accessor import WoodworkColumnAccessor
from woodwork.column_schema import ColumnSchema
from woodwork.exceptions import (
    ParametersIgnoredWarning,
    TypeConversionError,
    TypeValidationError,
    TypingInfoMismatchWarning,
    WoodworkNotInitError,
)
from woodwork.logical_types import (
    Categorical,
    CountryCode,
    CurrencyCode,
    Double,
    Integer,
    LatLong,
    NaturalLanguage,
    Ordinal,
    PostalCode,
    SubRegionCode,
)
from woodwork.tests.testing_utils import (
    is_property,
    is_public_method,
)


def test_accessor_init(sample_series):
    assert sample_series.ww._schema is None
    sample_series.ww.init()
    assert isinstance(sample_series.ww._schema, ColumnSchema)
    assert isinstance(sample_series.ww.logical_type, Categorical)
    assert sample_series.ww.semantic_tags == {"category"}


def test_accessor_init_with_schema(sample_series):
    sample_series.ww.init(
        semantic_tags={"test_tag"},
        description="this is a column",
        origin="base",
    )
    schema = sample_series.ww.schema

    head_series = sample_series.head(2)
    assert head_series.ww.schema is None
    head_series.ww.init(schema=schema)

    assert head_series.ww._schema is schema
    assert head_series.ww.description == "this is a column"
    assert head_series.ww.origin == "base"
    assert head_series.ww.semantic_tags == {"test_tag", "category"}

    iloc_series = sample_series.loc[[2, 3]]
    assert iloc_series.ww.schema is None
    iloc_series.ww.init(schema=schema)

    assert iloc_series.ww._schema is schema
    assert iloc_series.ww.description == "this is a column"
    assert iloc_series.ww.origin == "base"
    assert iloc_series.ww.semantic_tags == {"test_tag", "category"}
    assert isinstance(iloc_series.ww.logical_type, Categorical)


def test_accessor_init_with_schema_errors(sample_series):
    sample_series.ww.init(
        semantic_tags={"test_tag"},
        description="this is a column",
        origin="base",
    )
    schema = sample_series.ww.schema

    head_series = sample_series.head(2)
    assert head_series.ww.schema is None

    error = "Provided schema must be a Woodwork.ColumnSchema object."
    with pytest.raises(TypeError, match=error):
        head_series.ww.init(schema=int)

    ltype_dtype = "category"
    new_dtype = "string"

    error = re.escape(
        f"dtype mismatch between Series dtype {new_dtype}, and Categorical dtype, {ltype_dtype}",
    )
    diff_dtype_series = sample_series.astype(new_dtype)
    with pytest.raises(ValueError, match=error):
        diff_dtype_series.ww.init(schema=schema)


def test_accessor_with_schema_parameter_warning(sample_series):
    sample_series.ww.init(
        semantic_tags={"test_tag"},
        description="this is a column",
        origin="base",
    )
    schema = sample_series.ww.schema

    head_series = sample_series.head(2)

    warning = (
        "A schema was provided and the following parameters were ignored: "
        "logical_type, semantic_tags, description, origin, metadata, use_standard_tags"
    )
    with pytest.warns(ParametersIgnoredWarning, match=warning):
        head_series.ww.init(
            logical_type=Integer,
            semantic_tags={"ignored_test_tag"},
            description="an ignored description",
            origin="ignored origin",
            metadata={"user_id": "user1"},
            use_standard_tags=False,
            schema=schema,
        )

    assert head_series.ww.semantic_tags == {"category", "test_tag"}
    assert isinstance(head_series.ww.logical_type, Categorical)
    assert head_series.ww.description == "this is a column"
    assert head_series.ww.origin == "base"
    assert head_series.ww.metadata == {}


def test_accessor_init_with_logical_type(sample_series):
    series = sample_series.astype("string")
    series.ww.init(logical_type=NaturalLanguage)
    assert isinstance(series.ww.logical_type, NaturalLanguage)
    assert series.ww.semantic_tags == set()

    series = sample_series.astype("string")
    series.ww.init(logical_type="natural_language")
    assert isinstance(series.ww.logical_type, NaturalLanguage)
    assert series.ww.semantic_tags == set()

    series = sample_series.astype("string")
    series.ww.init(logical_type="NaturalLanguage")
    assert isinstance(series.ww.logical_type, NaturalLanguage)
    assert series.ww.semantic_tags == set()


def test_accessor_init_with_invalid_logical_type(sample_series):
    series_dtype = "object"
    series = sample_series.astype(series_dtype)
    correct_dtype = "string"
    error_message = re.escape(
        f"Cannot initialize Woodwork. Series dtype '{series_dtype}' is incompatible with "
        f"NaturalLanguage LogicalType. Try converting series dtype to '{correct_dtype}' before "
        "initializing or use the woodwork.init_series function to initialize.",
    )
    with pytest.raises(TypeValidationError, match=error_message):
        series.ww.init(logical_type=NaturalLanguage)


def test_accessor_init_with_semantic_tags(sample_series):
    semantic_tags = ["tag1", "tag2"]
    sample_series.ww.init(semantic_tags=semantic_tags, use_standard_tags=False)
    assert sample_series.ww.semantic_tags == set(semantic_tags)


def test_error_accessing_properties_before_init(sample_series):
    props_to_exclude = ["iloc", "loc", "schema", "_series"]
    props = [
        prop
        for prop in dir(sample_series.ww)
        if is_property(WoodworkColumnAccessor, prop) and prop not in props_to_exclude
    ]

    error = (
        "Woodwork not initialized for this Series. Initialize by calling Series.ww.init"
    )
    for prop in props:
        with pytest.raises(WoodworkNotInitError, match=error):
            getattr(sample_series.ww, prop)


def test_error_accessing_methods_before_init(sample_series):
    methods_to_exclude = ["init"]
    public_methods = [
        method
        for method in dir(sample_series.ww)
        if is_public_method(WoodworkColumnAccessor, method)
    ]
    public_methods = [
        method for method in public_methods if method not in methods_to_exclude
    ]

    method_args_dict = {
        "add_semantic_tags": [{"new_tag"}],
        "remove_semantic_tags": [{"new_tag"}],
        "set_logical_type": ["Integer"],
        "set_semantic_tags": [{"new_tag"}],
    }
    error = (
        "Woodwork not initialized for this Series. Initialize by calling Series.ww.init"
    )
    for method in public_methods:
        func = getattr(sample_series.ww, method)
        method_args = method_args_dict.get(method)
        with pytest.raises(WoodworkNotInitError, match=error):
            if method_args:
                func(*method_args)
            else:
                func()


def test_accessor_with_alternate_semantic_tags_input(sample_series):
    series = sample_series.copy()
    semantic_tags = "custom_tag"
    series.ww.init(semantic_tags=semantic_tags, use_standard_tags=False)
    assert series.ww.semantic_tags == {"custom_tag"}

    series = sample_series.copy()
    semantic_tags = {"custom_tag", "numeric"}
    series.ww.init(semantic_tags=semantic_tags, use_standard_tags=False)
    assert series.ww.semantic_tags == semantic_tags


def test_logical_type_errors(sample_series):
    error_message = "Invalid logical type specified for 'sample_series'"
    with pytest.raises(TypeError, match=error_message):
        sample_series.ww.init(logical_type=int)

    error_message = "String naturalllanguage is not a valid logical type"
    with pytest.raises(ValueError, match=error_message):
        sample_series.ww.init(logical_type="naturalllanguage")


def test_semantic_tag_errors(sample_series):
    error_message = "semantic_tags must be a string, set or list"
    with pytest.raises(TypeError, match=error_message):
        sample_series.ww.init(semantic_tags=int)

    error_message = "semantic_tags must be a string, set or list"
    with pytest.raises(TypeError, match=error_message):
        sample_series.ww.init(semantic_tags={"index": {}, "time_index": {}})

    error_message = "semantic_tags must contain only strings"
    with pytest.raises(TypeError, match=error_message):
        sample_series.ww.init(semantic_tags=["index", 1])


def test_accessor_description(sample_series):
    column_description = "custom description"
    sample_series.ww.init(description=column_description)
    assert sample_series.ww.description == column_description

    new_description = "updated description text"
    sample_series.ww.description = new_description
    assert sample_series.ww.description == new_description


def test_description_setter_error_before_init(sample_series):
    err_msg = (
        "Woodwork not initialized for this Series. Initialize by calling Series.ww.init"
    )
    with pytest.raises(WoodworkNotInitError, match=err_msg):
        sample_series.ww.description = "description"


def test_description_error_on_init(sample_series):
    err_msg = "Column description must be a string"
    with pytest.raises(TypeError, match=err_msg):
        sample_series.ww.init(description=123)


def test_description_error_on_update(sample_series):
    sample_series.ww.init()
    err_msg = "Column description must be a string"
    with pytest.raises(TypeError, match=err_msg):
        sample_series.ww.description = 123


def test_accessor_origin(sample_series):
    column_origin = "base"
    sample_series.ww.init(origin=column_origin)
    assert sample_series.ww.origin == column_origin

    new_origin = "engineered"
    sample_series.ww.origin = new_origin
    assert sample_series.ww.origin == new_origin


def test_origin_setter_error_before_init(sample_series):
    err_msg = (
        "Woodwork not initialized for this Series. Initialize by calling Series.ww.init"
    )
    with pytest.raises(WoodworkNotInitError, match=err_msg):
        sample_series.ww.origin = "base"


def test_origin_error_on_init(sample_series):
    err_msg = "Column origin must be a string"
    with pytest.raises(TypeError, match=err_msg):
        sample_series.ww.init(origin=123)


def test_origin_error_on_update(sample_series):
    sample_series.ww.init()
    err_msg = "Column origin must be a string"
    with pytest.raises(TypeError, match=err_msg):
        sample_series.ww.origin = 123


def test_accessor_repr(sample_series):
    sample_series.ww.init(use_standard_tags=False)
    dtype = "category"
    assert (
        sample_series.ww.__repr__()
        == f"<Series: sample_series (Physical Type = {dtype}) "
        "(Logical Type = Categorical) (Semantic Tags = set())>"
    )


def test_accessor_repr_error_before_init(sample_series):
    err_msg = (
        "Woodwork not initialized for this Series. Initialize by calling Series.ww.init"
    )
    with pytest.raises(WoodworkNotInitError, match=err_msg):
        sample_series.ww.__repr__()


def test_set_semantic_tags(sample_series):
    semantic_tags = {"tag1", "tag2"}
    sample_series.ww.init(semantic_tags=semantic_tags)
    assert sample_series.ww.semantic_tags == semantic_tags.union({"category"})

    new_tags = ["new_tag"]
    sample_series.ww.set_semantic_tags(new_tags)
    assert sample_series.ww.semantic_tags == set(new_tags).union({"category"})


def test_adds_numeric_standard_tag():
    series = pd.Series([1, 2, 3])
    semantic_tags = "custom_tag"

    logical_types = [Integer, Double]
    for logical_type in logical_types:
        series = series.astype(logical_type.primary_dtype)
        series.ww.init(logical_type=logical_type, semantic_tags=semantic_tags)
        assert series.ww.semantic_tags == {"custom_tag", "numeric"}


def test_adds_category_standard_tag():
    semantic_tags = "custom_tag"

    logical_types = [
        Categorical,
        CountryCode,
        CurrencyCode,
        Ordinal(order=(1, 2, 3)),
        PostalCode,
        SubRegionCode,
    ]
    for logical_type in logical_types:
        series = pd.Series([1, 2, 3], dtype="category")
        series.ww.init(logical_type=logical_type, semantic_tags=semantic_tags)
        assert series.ww.semantic_tags == {"custom_tag", "category"}


def test_does_not_add_standard_tags():
    series = pd.Series([1.1, 2, 3])
    semantic_tags = "custom_tag"
    series.ww.init(
        logical_type=Double,
        semantic_tags=semantic_tags,
        use_standard_tags=False,
    )
    assert series.ww.semantic_tags == {"custom_tag"}


def test_add_custom_tags(sample_series):
    semantic_tags = "initial_tag"
    sample_series.ww.init(semantic_tags=semantic_tags)

    sample_series.ww.add_semantic_tags("string_tag")
    assert sample_series.ww.semantic_tags == {"initial_tag", "string_tag", "category"}

    sample_series.ww.add_semantic_tags(["list_tag"])
    assert sample_series.ww.semantic_tags == {
        "initial_tag",
        "string_tag",
        "list_tag",
        "category",
    }

    sample_series.ww.add_semantic_tags({"set_tag"})
    assert sample_series.ww.semantic_tags == {
        "initial_tag",
        "string_tag",
        "list_tag",
        "set_tag",
        "category",
    }


def test_set_logical_type_without_standard_tags(sample_series):
    sample_series.ww.init(
        logical_type="Categorical",
        semantic_tags="original_tag",
        use_standard_tags=False,
    )

    new_series = sample_series.ww.set_logical_type("CountryCode")
    assert isinstance(sample_series.ww.logical_type, Categorical)
    assert isinstance(new_series.ww.logical_type, CountryCode)
    assert new_series.ww.semantic_tags == set()


def test_set_logical_type_valid_dtype_change(sample_series):
    sample_series.ww.init(logical_type="Categorical")

    new_series = sample_series.ww.set_logical_type("NaturalLanguage")

    original_dtype = "category"
    new_dtype = "string"

    assert isinstance(sample_series.ww.logical_type, Categorical)
    assert sample_series.dtype == original_dtype
    assert isinstance(new_series.ww.logical_type, NaturalLanguage)
    assert new_series.dtype == new_dtype


def test_set_logical_type_invalid_dtype_change(sample_series):
    sample_series.ww.init(logical_type="Categorical")
    error_message = (
        "Error converting datatype for sample_series from type category to "
        "type int64. Please confirm the underlying data is consistent with logical type Integer."
    )
    with pytest.raises(TypeConversionError, match=error_message):
        sample_series.ww.set_logical_type("Integer")


def test_reset_semantic_tags_without_standard_tags(sample_series):
    semantic_tags = "initial_tag"
    sample_series.ww.init(semantic_tags=semantic_tags, use_standard_tags=False)

    sample_series.ww.reset_semantic_tags()
    assert sample_series.ww.semantic_tags == set()


def test_remove_semantic_tags(sample_series):
    tags_to_remove = ["tag1", ["tag1"], {"tag1"}]

    for tag in tags_to_remove:
        series = sample_series.copy()
        series.ww.init(semantic_tags=["tag1", "tag2"], use_standard_tags=True)
        series.ww.remove_semantic_tags(tag)
        assert series.ww.semantic_tags == {"tag2", "category"}


def test_series_methods_on_accessor(sample_series):
    sample_series.ww.init()

    copied_series = sample_series.ww.copy()
    assert copied_series is not sample_series
    assert copied_series.ww.schema == sample_series.ww.schema
    pd.testing.assert_series_equal(sample_series, copied_series)


def test_series_methods_on_accessor_without_standard_tags(sample_series):
    sample_series.ww.init(use_standard_tags=False)

    copied_series = sample_series.ww.copy()
    assert copied_series is not sample_series
    assert copied_series.ww.schema == sample_series.ww.schema
    pd.testing.assert_series_equal(sample_series, copied_series)


def test_series_methods_on_accessor_returning_series_valid_schema(sample_series):
    sample_series.ww.init()

    replace_series = sample_series.ww.replace("a", "d")
    assert replace_series.ww._schema == sample_series.ww._schema
    assert replace_series.ww._schema is not sample_series.ww._schema
    pd.testing.assert_series_equal(replace_series, sample_series.replace("a", "d"))


def test_series_methods_on_accessor_dtype_mismatch(sample_df):
    ints_series = sample_df["id"]
    ints_series.ww.init()

    assert isinstance(ints_series.ww.logical_type, Integer)
    assert str(ints_series.dtype) == "int64"

    warning = (
        "Operation performed by astype has invalidated the Woodwork typing information:\n "
        "dtype mismatch between original dtype, int64, and returned dtype, Int64.\n "
        "Please initialize Woodwork with Series.ww.init"
    )
    with pytest.warns(TypingInfoMismatchWarning, match=warning):
        series = ints_series.ww.astype("Int64")
    assert series.ww._schema is None


def test_series_methods_on_accessor_inplace(sample_series):
    comparison_series = sample_series.copy()

    sample_series.ww.init()
    comparison_series.ww.init()

    val = sample_series.ww.pop(0)
    assert sample_series.ww._schema == comparison_series.ww._schema
    assert len(sample_series) == len(comparison_series) - 1
    assert val == "a"


def test_series_methods_on_accessor_returning_series_invalid_schema(sample_series):
    sample_series.ww.init()

    original_type = "category"
    new_type = "string"

    warning = (
        "Operation performed by astype has invalidated the Woodwork typing information:\n "
        f"dtype mismatch between original dtype, {original_type}, and returned dtype, {new_type}.\n "
        "Please initialize Woodwork with Series.ww.init"
    )

    with pytest.warns(TypingInfoMismatchWarning, match=warning):
        new_series = sample_series.ww.astype(new_type)

    assert new_series.ww._schema is None


def test_series_methods_on_accessor_other_returns(sample_series):
    sample_series.ww.init()
    col_shape = sample_series.ww.shape
    series_shape = sample_series.shape
    assert col_shape == (len(sample_series),)
    assert col_shape == series_shape

    assert sample_series.name == sample_series.ww.name
    series_nunique = sample_series.nunique()
    ww_nunique = sample_series.ww.nunique()
    assert series_nunique == ww_nunique


def test_series_methods_on_accessor_new_schema_object(sample_series):
    sample_series.ww.init(
        semantic_tags=["new_tag", "tag2"],
        metadata={"important_keys": [1, 2, 3]},
    )

    copied_series = sample_series.ww.copy()

    assert copied_series.ww._schema == sample_series.ww._schema
    assert copied_series.ww._schema is not sample_series.ww._schema

    copied_series.ww.metadata["important_keys"].append(4)
    assert copied_series.ww.metadata["important_keys"] == [1, 2, 3, 4]
    assert sample_series.ww.metadata["important_keys"] == [1, 2, 3]

    copied_series.ww.add_semantic_tags(["tag3"])
    assert copied_series.ww.semantic_tags == {"category", "new_tag", "tag2", "tag3"}
    assert sample_series.ww.semantic_tags == {"category", "new_tag", "tag2"}


def test_series_getattr_errors(sample_series):
    error_message = (
        "Woodwork not initialized for this Series. Initialize by calling Series.ww.init"
    )
    with pytest.raises(WoodworkNotInitError, match=error_message):
        sample_series.ww.shape

    sample_series.ww.init()
    error_message = "Woodwork has no attribute 'invalid_attr'"
    with pytest.raises(AttributeError, match=error_message):
        sample_series.ww.invalid_attr


def test_ordinal_requires_instance_on_init(sample_series):
    error_msg = "Must use an Ordinal instance with order values defined"
    with pytest.raises(TypeError, match=error_msg):
        sample_series.ww.init(logical_type=Ordinal)
    with pytest.raises(TypeError, match=error_msg):
        sample_series.ww.init(logical_type="Ordinal")


def test_ordinal_requires_instance_on_update(sample_series):
    sample_series.ww.init(logical_type="Categorical")

    error_msg = "Must use an Ordinal instance with order values defined"
    with pytest.raises(TypeError, match=error_msg):
        sample_series.ww.set_logical_type(Ordinal)
    with pytest.raises(TypeError, match=error_msg):
        sample_series.ww.set_logical_type("Ordinal")


def test_ordinal_with_order(sample_series):
    series = sample_series.copy()
    ordinal_with_order = Ordinal(order=["a", "b", "c"])
    series.ww.init(logical_type=ordinal_with_order)
    assert isinstance(series.ww.logical_type, Ordinal)
    assert series.ww.logical_type.order == ["a", "b", "c"]

    series = sample_series.copy()
    series.ww.init(logical_type="Categorical")
    new_series = series.ww.set_logical_type(ordinal_with_order)
    assert isinstance(new_series.ww.logical_type, Ordinal)
    assert new_series.ww.logical_type.order == ["a", "b", "c"]


def test_ordinal_with_incomplete_ranking(sample_series):
    ordinal_incomplete_order = Ordinal(order=["a", "b"])
    error_msg = re.escape(
        "Ordinal column sample_series contains values that are not "
        "present in the order values provided: ['c']",
    )
    with pytest.raises(ValueError, match=error_msg):
        sample_series.ww.init(logical_type=ordinal_incomplete_order)


def test_ordinal_with_nan_values():
    nan_series = pd.Series(["a", "b", np.nan, "a"]).astype("category")
    ordinal_with_order = Ordinal(order=["a", "b"])
    nan_series.ww.init(logical_type=ordinal_with_order)
    assert isinstance(nan_series.ww.logical_type, Ordinal)
    assert nan_series.ww.logical_type.order == ["a", "b"]


def test_latlong_init_with_valid_series(latlongs):
    series = latlongs[0]
    series.ww.init(logical_type="LatLong")
    assert isinstance(series.ww.logical_type, LatLong)


def test_latlong_init_error_with_invalid_series(latlongs):
    series = latlongs[1]
    error_message = (
        "Cannot initialize Woodwork. Series does not contain properly formatted "
        "LatLong data. Try reformatting before initializing or use the "
        "woodwork.init_series function to initialize."
    )
    with pytest.raises(TypeValidationError, match=error_message):
        series.ww.init(logical_type="LatLong")


def test_latlong_formatting_with_init_series(latlongs):
    expected_series = pd.Series([(1.0, 2.0), (3.0, 4.0)])

    expected_series.ww.init(logical_type=LatLong)
    for series in latlongs:
        new_series = init_series(series, logical_type=LatLong)
        assert isinstance(new_series.ww.logical_type, LatLong)
        pd.testing.assert_series_equal(new_series, expected_series)
        assert expected_series.ww._schema == new_series.ww._schema


def test_accessor_equality(sample_series):
    # Check different parameters
    str_col = sample_series.copy()
    str_col.ww.init(logical_type="Categorical")

    str_col_2 = sample_series.copy()
    str_col_2.ww.init(logical_type=Categorical)

    str_col_diff_tags = sample_series.copy()
    str_col_diff_tags.ww.init(logical_type=Categorical, semantic_tags={"test"})

    diff_name_col = sample_series.copy()
    diff_name_col.name = "different_name"
    diff_name_col.ww.init(logical_type=Categorical)

    diff_dtype_col = sample_series.astype("string")
    diff_dtype_col.ww.init(logical_type=NaturalLanguage)

    assert str_col.ww == str_col_2.ww
    assert str_col.ww != str_col_diff_tags.ww
    assert str_col.ww != diff_name_col.ww
    assert str_col.ww != diff_dtype_col.ww

    # Check different underlying series
    str_col = sample_series.astype("string")
    str_col.ww.init(logical_type="NaturalLanguage")
    changed_series = sample_series.copy().replace(to_replace="a", value="test")
    changed_series = changed_series.astype("string")
    changed_series.ww.init(logical_type="NaturalLanguage")

    assert str_col.ww != changed_series.ww


def test_accessor_shallow_equality(sample_series):
    metadata_col = init_series(
        sample_series.copy(),
        logical_type="NaturalLanguage",
        metadata={"interesting_values": ["a", "b"]},
    )
    diff_metadata_col = init_series(
        sample_series.copy(),
        logical_type="NaturalLanguage",
        metadata={"interesting_values": ["c"]},
    )

    assert metadata_col.ww.__eq__(diff_metadata_col.ww, deep=False)
    assert not metadata_col.ww.__eq__(diff_metadata_col.ww, deep=True)

    schema = metadata_col.ww.schema
    diff_data_col = metadata_col.replace(to_replace="a", value="1")
    # dtype gets changed to object in replace
    diff_data_col = diff_data_col.astype("string")

    diff_data_col.ww.init(schema=schema)
    same_data_col = metadata_col.ww.copy()

    assert diff_data_col.ww.schema.__eq__(metadata_col.ww.schema, deep=True)
    assert same_data_col.ww.schema.__eq__(metadata_col.ww.schema, deep=True)

    assert diff_data_col.ww.__eq__(metadata_col.ww, deep=False)
    assert same_data_col.ww.__eq__(metadata_col.ww, deep=False)
    assert same_data_col.ww.__eq__(metadata_col.ww, deep=True)
    assert not diff_data_col.ww.__eq__(metadata_col.ww, deep=True)


def test_accessor_metadata(sample_series):
    column_metadata = {"metadata_field": [1, 2, 3], "created_by": "user0"}

    series = sample_series.copy()
    series.ww.init()
    assert series.ww.metadata == {}

    series = sample_series.copy()
    series.ww.init(metadata=column_metadata)
    assert series.ww.metadata == column_metadata

    new_metadata = {"date_created": "1/1/19", "created_by": "user1"}

    series.ww.metadata = {**series.ww.metadata, **new_metadata}
    assert series.ww.metadata == {
        "date_created": "1/1/19",
        "metadata_field": [1, 2, 3],
        "created_by": "user1",
    }

    series.ww.metadata.pop("created_by")
    assert series.ww.metadata == {"date_created": "1/1/19", "metadata_field": [1, 2, 3]}

    series.ww.metadata["number"] = 1012034
    assert series.ww.metadata == {
        "date_created": "1/1/19",
        "metadata_field": [1, 2, 3],
        "number": 1012034,
    }


def test_metadata_setter_error_before_init(sample_series):
    err_msg = (
        "Woodwork not initialized for this Series. Initialize by calling Series.ww.init"
    )
    with pytest.raises(WoodworkNotInitError, match=err_msg):
        sample_series.ww.metadata = {"key": "val"}


def test_accessor_metadata_error_on_init(sample_series):
    err_msg = "Column metadata must be a dictionary"
    with pytest.raises(TypeError, match=err_msg):
        sample_series.ww.init(metadata=123)


def test_accessor_metadata_error_on_update(sample_series):
    sample_series.ww.init()
    err_msg = "Column metadata must be a dictionary"
    with pytest.raises(TypeError, match=err_msg):
        sample_series.ww.metadata = 123


def test_non_string_column_name(sample_series):
    sample_series.name = 0
    sample_series.ww.init(semantic_tags={"test_tag"})

    assert sample_series.ww.name == 0
    assert sample_series.name == 0
    assert sample_series.ww.semantic_tags == {"category", "test_tag"}


def test_schema_property(sample_series):
    sample_series.ww.init(
        logical_type="Categorical",
        semantic_tags="test_tag",
        metadata={"created_by": "user1"},
    )

    assert sample_series.ww.schema is not sample_series.ww._schema
    assert sample_series.ww.schema == sample_series.ww._schema

    changed_schema = sample_series.ww.schema
    changed_schema.metadata["new"] = 1

    assert changed_schema != sample_series.ww._schema


@patch("woodwork.logical_types.LogicalType.validate")
def test_validation_methods_called_init(mock_validate, sample_series):
    assert not mock_validate.called

    not_validated = sample_series.copy()
    not_validated.ww.init(validate=False)

    assert not mock_validate.called

    validated = sample_series.copy()
    validated.ww.init(validate=True)

    assert not mock_validate.called
    assert validated.ww == not_validated.ww
    pd.testing.assert_series_equal(validated, not_validated)


@patch("woodwork.logical_types.LogicalType.validate")
def test_ordinal_validation_methods_called_init(mock_validate, sample_series):
    assert not mock_validate.called

    not_validated = sample_series.copy()
    not_validated.ww.init(logical_type=Ordinal(order=["a", "b", "c"]), validate=False)

    assert not mock_validate.called

    validated = sample_series.copy()
    validated.ww.init(logical_type=Ordinal(order=["a", "b", "c"]), validate=True)

    assert mock_validate.called
    assert validated.ww == not_validated.ww
    pd.testing.assert_series_equal(validated, not_validated)


@patch("woodwork.logical_types.LogicalType.validate")
def test_latlong_validation_methods_called_init(mock_validate, latlong_df):
    assert not mock_validate.called

    latlong_series = latlong_df["null_latitude"]
    not_validated = latlong_series.copy()
    not_validated.ww.init(LatLong, validate=False)

    assert not mock_validate.called

    validated = latlong_series.copy()
    validated.ww.init(LatLong, validate=True)

    assert mock_validate.called
    assert validated.ww == not_validated.ww
    pd.testing.assert_series_equal(validated, not_validated)


@patch("woodwork.column_accessor._validate_schema")
def test_validation_methods_called_init_with_schema(
    mock_validate_schema,
    sample_series,
):
    assert not mock_validate_schema.called
    schema_series = sample_series.copy()
    schema_series.ww.init()
    schema = schema_series.ww.schema

    not_validated = sample_series.copy()
    not_validated.ww.init(schema=schema, validate=False)

    assert not mock_validate_schema.called

    validated = sample_series.copy()
    validated.ww.init(schema=schema, validate=True)

    assert mock_validate_schema.called
    assert validated.ww == not_validated.ww
    pd.testing.assert_series_equal(validated, not_validated)


def test_series_methods_returning_frame(sample_series):
    sample_series.ww.init(
        semantic_tags={"test_tag"},
        description="custom description",
        origin="base",
        metadata={"custom key": "custom value"},
    )
    sample_frame = sample_series.ww.to_frame()

    assert _is_dataframe(sample_frame)
    assert sample_frame.ww.schema is not None
    assert sample_frame.ww.columns["sample_series"] == sample_series.ww.schema

    reset_index_frame = sample_series.ww.reset_index(drop=False)
    assert _is_dataframe(reset_index_frame)
    assert reset_index_frame.ww.schema is not None


def test_series_methods_returning_frame_no_name(sample_series):
    sample_series.ww.init(
        semantic_tags={"test_tag"},
        description="custom description",
        origin="base",
        metadata={"custom key": "custom value"},
    )
    sample_series.name = None
    assert sample_series.name is None
    sample_series.ww.name = None
    assert sample_series.ww.name is None
    sample_frame = sample_series.ww.to_frame()

    assert _is_dataframe(sample_frame)
    assert sample_frame.ww.schema is not None
    assert sample_frame.ww.columns[0] == sample_series.ww.schema

    reset_index_frame = sample_series.ww.reset_index(drop=False)
    assert _is_dataframe(reset_index_frame)
    assert reset_index_frame.ww.schema is not None


EXPECTED_COLUMN_NULLABILITIES = {
    "id": False,
    "full_name": True,
    "email": True,
    "phone_number": True,
    "age": True,
    "signup_date": True,
    "is_registered": True,
    "double": True,
    "double_with_nan": True,
    "integer": False,
    "nullable_integer": True,
    "boolean": False,
    "categorical": True,
    "datetime_with_NaT": True,
    "url": True,
    "ip_address": True,
}


def test_nullable_attribute(sample_df):
    sample_df.ww.init()

    for key in sample_df.ww.columns:
        actual = sample_df.ww[key].ww.nullable
        expected = EXPECTED_COLUMN_NULLABILITIES[key]

        assert actual is expected


def test_validate_logical_type(sample_df):
    series = sample_df["email"]
    series = init_series(series, logical_type="EmailAddress")
    assert series.ww.validate_logical_type() is None

    invalid_row = pd.Series({4: "bad_email"}, name="email", dtype="string")

    series = pd.concat([series, invalid_row])
    series = init_series(series, logical_type="EmailAddress")
    match = "Series email contains invalid email address values. "
    match += "The email_inference_regex can be changed in the config if needed."

    with pytest.raises(TypeValidationError, match=match):
        series.ww.validate_logical_type()

    actual = series.ww.validate_logical_type(return_invalid_values=True)
    expected = pd.Series({4: "bad_email"}, dtype="string")
    assert actual.equals(expected)



import warnings
from datetime import datetime
from typing import Optional

import numpy as np
import pandas as pd
from pandas import CategoricalDtype
from pandas.api import types as pdtypes

import woodwork as ww
from woodwork.config import config
from woodwork.exceptions import (
    TypeConversionError,
    TypeConversionWarning,
    TypeValidationError,
)
from woodwork.type_sys.utils import _get_specified_ltype_params
from woodwork.utils import (
    _infer_datetime_format,
    _is_valid_latlong_series,
    _is_valid_latlong_value,
    _reformat_to_latlong,
    camel_to_snake,
)


class ClassNameDescriptor(object):
    """Descriptor to convert a class's name from camelcase to snakecase"""

    def __get__(self, instance, class_):
        return camel_to_snake(class_.__name__)


class LogicalTypeMetaClass(type):
    def __repr__(cls):
        return cls.__name__


class LogicalType(object, metaclass=LogicalTypeMetaClass):
    """Base class for all other Logical Types"""

    type_string = ClassNameDescriptor()
    primary_dtype = "string"
    standard_tags = set()

    def __eq__(self, other, deep=False):
        return isinstance(other, self.__class__) and _get_specified_ltype_params(
            other,
        ) == _get_specified_ltype_params(self)

    def __str__(self):
        return str(self.__class__)

    @classmethod
    def _get_valid_dtype(cls, series_type):
        """Return the dtype that is considered valid for a series with the given logical_type"""
        return cls.primary_dtype

    def transform(self, series, null_invalid_values=False):
        """Converts the series dtype to match the logical type's if it is different."""
        new_dtype = self._get_valid_dtype(type(series))
        if new_dtype != str(series.dtype):
            # Update the underlying series
            try:
                series = series.astype(new_dtype)
            except (TypeError, ValueError):
                raise TypeConversionError(series, new_dtype, type(self))
        return series

    def validate(self, series, *args, **kwargs):
        """Validates that a logical type is consistent with the series dtype. Performs additional type
        specific validation, as required. When the series' dtype does not match the logical types' required dtype,
        raises a TypeValidationError."""
        valid_dtype = self._get_valid_dtype(type(series))
        if valid_dtype != str(series.dtype):
            raise TypeValidationError(
                f"Series dtype '{series.dtype}' is incompatible with {self.type_string} LogicalType, try converting to {valid_dtype} dtype",
            )


class Address(LogicalType):
    """Represents Logical Types that contain address values.

    Examples:
        .. code-block:: python

            ['1 Miller Drive, New York, NY 12345', '1 Berkeley Street, Boston, MA 67891']
            ['26387 Russell Hill, Dallas, TX 34521', '54305 Oxford Street, Seattle, WA 95132']
    """

    primary_dtype = "string"


class Age(LogicalType):
    """Represents Logical Types that contain whole numbers indicating a person's age.
    Has 'numeric' as a standard tag.

    Examples:
        .. code-block:: python

            [15, 22, 45]
            [30, 62, 87]
    """

    primary_dtype = "int64"
    standard_tags = {"numeric"}

    def validate(self, series, return_invalid_values=True):
        """Validates age values by checking for non-negative values.

        Args:
            series (Series): Series of age values
            return_invalid_values (bool): Whether or not to return invalid age values

        Returns:
            Series: If return_invalid_values is True, returns invalid age values.
        """
        return _validate_age(series, return_invalid_values)


class AgeFractional(LogicalType):
    """Represents Logical Types that contain non-negative floating point numbers indicating a person's age.
    Has 'numeric' as a standard tag. May also contain null values.

    Examples:
        .. code-block:: python

            [0.34, 24.34, 45.0]
            [30.5, 62.82, np.nan]
    """

    primary_dtype = "float64"
    standard_tags = {"numeric"}

    def transform(self, series, null_invalid_values=False):
        if null_invalid_values:
            series = _coerce_age(series, fractional=True)
        return super().transform(series)

    def validate(self, series, return_invalid_values=True):
        """Validates age values by checking for non-negative values.

        Args:
            series (Series): Series of age values
            return_invalid_values (bool): Whether or not to return invalid age values

        Returns:
            Series: If return_invalid_values is True, returns invalid age values.
        """
        return _validate_age(series, return_invalid_values)


class AgeNullable(LogicalType):
    """Represents Logical Types that contain whole numbers indicating a person's age.
    Has 'numeric' as a standard tag. May also contain null values.

    Examples:
        .. code-block:: python

            [np.nan, 22, 45]
            [30, 62, np.nan]
    """

    primary_dtype = "Int64"
    standard_tags = {"numeric"}

    def transform(self, series, null_invalid_values=False):
        if null_invalid_values:
            series = _coerce_age(series, fractional=False)
        return super().transform(series)

    def validate(self, series, return_invalid_values=True):
        """Validates age values by checking for non-negative values.

        Args:
            series (Series): Series of age values
            return_invalid_values (bool): Whether or not to return invalid age values

        Returns:
            Series: If return_invalid_values is True, returns invalid age values.
        """
        return _validate_age(series, return_invalid_values)


class Boolean(LogicalType):
    """Represents Logical Types that contain binary values indicating true/false.

    Args:
        cast_nulls_as (bool): If provided, null values in the column will be cast to this default bool, otherwise will raise an error if None.
            Defaults to None.

    Examples:
        .. code-block:: python

            [True, False, True]
            [0, 1, 1]
    """

    primary_dtype = "bool"

    def __init__(self, cast_nulls_as=None):
        if cast_nulls_as and not isinstance(cast_nulls_as, bool):
            raise ValueError(
                f"Parameter `cast_nulls_as` must be either True or False, recieved {cast_nulls_as}",
            )
        self.cast_nulls_as = cast_nulls_as

    def transform(self, series, null_invalid_values=False):
        """Validates Boolean values by checking for valid boolean equivalents.

        Args:
            series (series): Series of boolean values

        Returns:
            Series: Returns column transformed into boolean type
        """
        ve = ValueError(
            "Expected no null values in this Boolean column. If you want to keep the nulls, use BooleanNullable type. Otherwise, cast these nulls to a boolean value with the `cast_null_as` parameter.",
        )
        if not pdtypes.is_dtype_equal("bool", series.dtype):
            if series.isna().any():
                if self.cast_nulls_as is None:
                    raise ve
                series.fillna(self.cast_nulls_as, inplace=True)
            series = _coerce_boolean(series, True)
        return super().transform(series)


class BooleanNullable(LogicalType):
    """Represents Logical Types that contain binary values indicating true/false.
    May also contain null values.

    Examples:
        .. code-block:: python

            [True, False, None]
            [0, 1, 1]
    """

    primary_dtype = "boolean"

    def transform(self, series, null_invalid_values=False):
        series = _replace_nans(series, self.primary_dtype)
        series = _coerce_boolean(series, null_invalid_values)
        return super().transform(series)


class Categorical(LogicalType):
    """Represents Logical Types that contain unordered discrete values that fall
    into one of a set of possible values. Has 'category' as a standard tag.

    Examples:
        .. code-block:: python

            ["red", "green", "blue"]
            ["produce", "dairy", "bakery"]
            [3, 1, 2]
    """

    primary_dtype = "category"
    standard_tags = {"category"}

    def __init__(self, encoding=None):
        # encoding dict(str -> int)
        # user can specify the encoding to use downstream
        pass


class CountryCode(LogicalType):
    """Represents Logical Types that use the ISO-3166 standard country code to represent countries.
    ISO 3166-1 (countries) are supported. These codes should be in the Alpha-2 format.

    Examples:
        .. code-block:: python

            ["AU", "US", "UA"]
            ["GB", "NZ", "DE"]
    """

    primary_dtype = "category"
    standard_tags = {"category"}


class CurrencyCode(LogicalType):
    """Represents Logical Types that use the ISO-4217 internation standard currency code to represent currencies.

    Examples:
        .. code-block:: python

            ["GBP", "JPY", "USD"]
            ["SAR", "EUR", "CZK"]
    """

    primary_dtype = "category"
    standard_tags = {"category"}


class Datetime(LogicalType):
    """Represents Logical Types that contain date and time information.

    Args:
        datetime_format (str): Desired datetime format for data

    Examples:
        .. code-block:: python

            ["2020-09-10",
             "2020-01-10 00:00:00",
             "01/01/2000 08:30"]
    """

    primary_dtype = "datetime64[ns]"
    datetime_format = None

    def __init__(self, datetime_format=None, timezone=None):
        self.datetime_format = datetime_format
        self.timezone = timezone

    def _remove_timezone(self, series):
        """Removes timezone from series and stores in logical type."""
        if hasattr(series.dtype, "tz") and series.dtype.tz:
            self.timezone = str(series.dtype.tz)
            series = series.dt.tz_localize(None)
        return series

    def transform(self, series, null_invalid_values=False):
        """Converts the series data to a formatted datetime. Datetime format will be inferred if datetime_format is None."""

        def _year_filter(date):
            """Applies a filter to the years to ensure that the pivot point isn't too far forward."""
            if date.year > datetime.today().year + 10:
                date = date.replace(year=date.year - 100)
            return date

        new_dtype = self._get_valid_dtype(type(series))
        series = self._remove_timezone(series)
        series_dtype = str(series.dtype)

        if new_dtype != series_dtype:
            self.datetime_format = self.datetime_format or _infer_datetime_format(
                series,
            )
            utc = self.datetime_format and self.datetime_format.endswith("%z")

            try:
                series = pd.to_datetime(
                    series,
                    format=self.datetime_format,
                    utc=utc,
                )
            except (TypeError, ValueError):
                warnings.warn(
                    f"Some rows in series '{series.name}' are incompatible with datetime format "
                    f"'{self.datetime_format}' and have been replaced with null values. You may be "
                    "able to fix this by using an instantiated Datetime logical type with a different format "
                    "string specified for this column during Woodwork initialization.",
                    TypeConversionWarning,
                )
                series = pd.to_datetime(
                    series,
                    format=self.datetime_format,
                    errors="coerce",
                    utc=utc,
                )

        series = self._remove_timezone(series)
        if self.datetime_format is not None and "%y" in self.datetime_format:
            series = series.apply(_year_filter)
        return super().transform(series)


class Double(LogicalType):
    """Represents Logical Types that contain positive and negative numbers, some of
    which include a fractional component. Includes zero (0).
    Has 'numeric' as a standard tag.

    Examples:
        .. code-block:: python

            [1.2, 100.4, 3.5]
            [-15.34, 100, 58.3]
    """

    primary_dtype = "float64"
    standard_tags = {"numeric"}

    def transform(self, series, null_invalid_values=False):
        series = _replace_nans(series, self.primary_dtype)
        if null_invalid_values:
            series = _coerce_numeric(series)
        return super().transform(series)


class Integer(LogicalType):
    """Represents Logical Types that contain positive and negative numbers
    without a fractional component, including zero (0).
    Has 'numeric' as a standard tag.

    Examples:
        .. code-block:: python

            [100, 35, 0]
            [-54, 73, 11]
    """

    primary_dtype = "int64"
    standard_tags = {"numeric"}


class IntegerNullable(LogicalType):
    """Represents Logical Types that contain positive and negative numbers
    without a fractional component, including zero (0). May contain null
    values. Has 'numeric' as a standard tag.

    Examples:
        .. code-block:: python

            [100, 35, np.nan]
            [-54, 73, 11]
    """

    primary_dtype = "Int64"
    standard_tags = {"numeric"}

    def transform(self, series, null_invalid_values=False):
        """Converts a series dtype to Int64.

        Args:
            series (Series): A series of data values.
            null_invalid_values (bool): If true, nulls invalid integers by coercing the series
                to string, numeric, and then nulling out floats with decimals. Defaults to False.

        Returns:
            Series: A series of integers.
        """
        series = _replace_nans(series, self.primary_dtype)
        if null_invalid_values:
            series = _coerce_integer(series)
        return super().transform(series)


class EmailAddress(LogicalType):
    """Represents Logical Types that contain email address values.

    Examples:
        .. code-block:: python

            ["john.smith@example.com",
             "support@example.com",
             "team@example.com"]
    """

    primary_dtype = "string"

    def transform(self, series, null_invalid_values=False):
        if null_invalid_values:
            series = _coerce_string(series, regex="email_inference_regex")
        return super().transform(series)

    def validate(self, series, return_invalid_values=False):
        """Validates email address values based on the regex in the config.

        Args:
            series (Series): Series of email address values
            return_invalid_values (bool): Whether or not to return invalid email address values

        Returns:
            Series: If return_invalid_values is True, returns invalid email address.
        """
        return _regex_validate("email_inference_regex", series, return_invalid_values)


class Filepath(LogicalType):
    """Represents Logical Types that specify locations of directories and files
    in a file system.

    Examples:
        .. code-block:: python

            ["/usr/local/bin",
             "/Users/john.smith/dev/index.html",
             "/tmp"]
    """

    primary_dtype = "string"


class PersonFullName(LogicalType):
    """Represents Logical Types that may contain first, middle and last names,
    including honorifics and suffixes.

    Examples:
        .. code-block:: python

            ["Mr. John Doe, Jr.",
             "Doe, Mrs. Jane",
             "James Brown"]
    """

    primary_dtype = "string"


class IPAddress(LogicalType):
    """Represents Logical Types that contain IP addresses, including both
    IPv4 and IPv6 addresses.

    Examples:
        .. code-block:: python

            ["172.16.254.1",
             "192.0.0.0",
             "2001:0db8:0000:0000:0000:ff00:0042:8329"]
    """

    primary_dtype = "string"


class LatLong(LogicalType):
    """Represents Logical Types that contain latitude and longitude values in decimal degrees.

    Note:
        LatLong values will be stored with the object dtype as a
        tuple of floats and must contain only two values.

        Null latitude or longitude values will be stored as np.nan, and
        a fully null LatLong (np.nan, np.nan) will be stored as just a
        single nan.

    Examples:
        .. code-block:: python

            [(33.670914, -117.841501),
             (40.423599, -86.921162),
             (-45.031705, nan)]
    """

    primary_dtype = "object"

    def transform(self, series, null_invalid_values=False):
        """Formats a series to be a tuple of two floats."""
        if null_invalid_values:
            series = _coerce_latlong(series)
        series = series.apply(_reformat_to_latlong)

        return super().transform(series)

    def validate(self, series, return_invalid_values=False):
        # TODO: we'll want to actually handle return_invalid_values in the ordinal and latlong logical types.
        super().validate(series)
        if not _is_valid_latlong_series(series):
            raise TypeValidationError(
                "Cannot initialize Woodwork. Series does not contain properly formatted "
                "LatLong data. Try reformatting before initializing or use the "
                "woodwork.init_series function to initialize.",
            )


class NaturalLanguage(LogicalType):
    """Represents Logical Types that contain text or characters representing
    natural human language

    Examples:
        .. code-block:: python

            ["This is a short sentence.",
             "I like to eat pizza!",
             "When will humans go to mars?"]
    """

    primary_dtype = "string"


class Unknown(LogicalType):
    """Represents Logical Types that cannot be inferred as a specific Logical Type. It is assumed to contain string data.

    Examples:
        .. code-block:: python

            ["ax23n9ck23l",
             "1,28&*_%*&&xejc",
             "xnmvz@@Dcmeods-0"]

    """

    primary_dtype = "string"


class Ordinal(LogicalType):
    """Represents Logical Types that contain ordered discrete values.
    Has 'category' as a standard tag.

    Args:
        order (list or tuple): An list or tuple specifying the order of the ordinal
            values from low to high. The underlying series cannot contain values that
            are not present in the order values.

    Examples:
        .. code-block:: python

            ["first", "second", "third"]
            ["bronze", "silver", "gold"]
    """

    primary_dtype = "category"
    standard_tags = {"category"}

    def __init__(self, order=None):
        self.order = order

    def _validate_order_values(self, series):
        """Make sure order values are properly defined and confirm the supplied series
        does not contain any values that are not in the specified order values"""
        if self.order is None:
            raise TypeError("Must use an Ordinal instance with order values defined")
        elif not isinstance(self.order, (list, tuple)):
            raise TypeError("Order values must be specified in a list or tuple")
        if len(self.order) != len(set(self.order)):
            raise ValueError("Order values cannot contain duplicates")

        if isinstance(series, pd.Series):
            missing_order_vals = set(series.dropna().values).difference(self.order)
            if missing_order_vals:
                error_msg = (
                    f"Ordinal column {series.name} contains values that are not present "
                    f"in the order values provided: {sorted(list(missing_order_vals))}"
                )
                raise ValueError(error_msg)

    def transform(self, series, null_invalid_values=False):
        """Validates the series and converts the dtype to match the logical type's if it is different."""
        self._validate_order_values(series)

        typed_ser = super().transform(series)
        if isinstance(typed_ser.dtype, CategoricalDtype):
            typed_ser = typed_ser.cat.set_categories(self.order, ordered=True)

        return typed_ser

    def validate(self, series, return_invalid_values=False):
        # TODO: we'll want to actually handle return_invalid_values in the ordinal and latlong logical types.
        super().validate(series)
        self._validate_order_values(series)

    def __str__(self):
        return "{}: {}".format(self.__class__, self.order)


class PhoneNumber(LogicalType):
    """Represents Logical Types that contain numeric digits and characters
    representing a phone number.

    Examples:
        .. code-block:: python

            ["1-(555)-123-5495",
             "+1-555-123-5495",
             "5551235495"]
    """

    primary_dtype = "string"

    def transform(self, series, null_invalid_values=False):
        if null_invalid_values:
            series = _coerce_string(series, regex="phone_inference_regex")
        return super().transform(series)

    def validate(self, series, return_invalid_values=False):
        """Validates PhoneNumber values based on the regex in the config.
        By default, this validates US/Canada-based phone numbers.

        Args:
            series (Series): Series of phone number values.
            return_invalid_values (bool): Whether or not to return invalid phone numbers.

        Returns:
            Series: If return_invalid_values is True, returns invalid phone numbers.
        """
        return _regex_validate("phone_inference_regex", series, return_invalid_values)


class SubRegionCode(LogicalType):
    """Represents Logical Types that use the ISO-3166 standard sub-region code to
    represent a portion of a larger geographic region. ISO 3166-2 (sub-regions)
    codes are supported. These codes should be in the Alpha-2 format.

    Examples:
        .. code-block:: python

            ["US-CO", "US-MA", "US-CA"]
            ["AU-NSW", "AU-TAS", "AU-QLD"]
    """

    primary_dtype = "category"
    standard_tags = {"category"}


class Timedelta(LogicalType):
    """Represents Logical Types that contain values specifying a duration of time

    Examples:
        .. code-block:: python

            [pd.Timedelta('1 days 00:00:00'),
             pd.Timedelta('-1 days +23:40:00'),
             pd.Timedelta('4 days 12:00:00')]
    """

    primary_dtype = "timedelta64[ns]"


class URL(LogicalType):
    """Represents Logical Types that contain URLs, which may include protocol, hostname
    and file name

    Examples:
        .. code-block:: python

            ["http://google.com",
             "https://example.com/index.html",
             "example.com"]
    """

    primary_dtype = "string"

    def transform(self, series, null_invalid_values=False):
        if null_invalid_values:
            series = _coerce_string(series, regex="url_inference_regex")
        return super().transform(series)

    def validate(self, series, return_invalid_values=False):
        """Validates URL values based on the regex in the config.

        Args:
            series (Series): Series of URL values
            return_invalid_values (bool): Whether or not to return invalid URLs

        Returns:
            Series: If return_invalid_values is True, returns invalid URLs.
        """
        return _regex_validate("url_inference_regex", series, return_invalid_values)


class PostalCode(LogicalType):
    """Represents Logical Types that contain a series of postal codes for
    representing a group of addresses. Has 'category' as a standard tag.

    Examples:
        .. code-block:: python

            ["90210"
             "60018-0123",
             "10010"]
    """

    primary_dtype = "category"
    standard_tags = {"category"}

    def transform(self, series, null_invalid_values=False):
        if null_invalid_values:
            series = _coerce_postal_code(series)

        if pd.api.types.is_numeric_dtype(series):
            try:
                series = series.astype("Int64").astype("string")
            except TypeError:
                raise TypeConversionError(series, "string", type(self))

        return super().transform(series)

    def validate(self, series, return_invalid_values=False):
        """Validates PostalCode values based on the regex in the config. Currently only validates US Postal codes.

        Args:
            series (Series): Series of PostalCode values.
            return_invalid_values (bool): Whether or not to return invalid PostalCodes.

        Returns:
            Series: If return_invalid_values is True, returns invalid PostalCodes.
        """
        return _regex_validate(
            "postal_code_inference_regex",
            series,
            return_invalid_values,
        )


_NULLABLE_PHYSICAL_TYPES = {
    "boolean",
    "category",
    "datetime64[ns]",
    "Int8",
    "Int16",
    "Int32",
    "Int64",
    "Float32",
    "Float64",
    "float16",
    "float32",
    "float64",
    "float128",
    "object",
    "string",
    "timedelta64[ns]",
}


def _regex_validate(regex_key, series, return_invalid_values):
    """Validates data values based on the logical type regex in the config."""
    invalid = _get_index_invalid_string(series, regex_key)

    if return_invalid_values:
        return series[invalid]

    else:
        any_invalid = invalid.any()

        if any_invalid:
            type_string = {
                "url_inference_regex": "url",
                "email_inference_regex": "email address",
                "phone_inference_regex": "phone number",
                "postal_code_inference_regex": "postal code",
            }[regex_key]

            info = f"Series {series.name} contains invalid {type_string} values. "
            info += f"The {regex_key} can be changed in the config if needed."
            raise TypeValidationError(info)


def _replace_nans(series: pd.Series, primary_dtype: Optional[str] = None) -> pd.Series:
    """
    Replaces empty string values, string representations of NaN values ("nan", "<NA>"), and NaN equivalents
    with np.nan or pd.NA depending on column dtype.
    """
    original_dtype = series.dtype
    if primary_dtype == str(original_dtype):
        return series
    if str(original_dtype) == "string":
        series = series.replace(ww.config.get_option("nan_values"), pd.NA)
        return series
    series = series.replace(ww.config.get_option("nan_values"), np.nan)
    if str(original_dtype) == "boolean":
        series = series.astype(original_dtype)

    return series


def _validate_age(series, return_invalid_values):
    """Validates data values are non-negative."""
    invalid = _get_index_invalid_age(series)
    if return_invalid_values:
        return series[invalid]

    else:
        any_invalid = invalid.any()

        if any_invalid:
            info = f"Series {series.name} contains negative values."
            raise TypeValidationError(info)


def _get_index_invalid_integer(series):
    return series.mod(1).ne(0)


def _get_index_invalid_string(series, regex_key):
    regex = config.get_option(regex_key)
    return ~series.str.match(regex).astype("boolean")


def _get_index_invalid_age(series):
    return series.lt(0)


def _get_index_invalid_latlong(series):
    return ~series.apply(_is_valid_latlong_value)


def _coerce_string(series, regex=None):
    if pd.api.types.is_object_dtype(series) or not pd.api.types.is_string_dtype(series):
        series = series.astype("string")

    if isinstance(regex, str):
        invalid = _get_index_invalid_string(series, regex)

        if invalid.any():
            series[invalid] = pd.NA
    return series


def _coerce_numeric(series):
    if not pd.api.types.is_numeric_dtype(series):
        series = pd.to_numeric(_coerce_string(series), errors="coerce")
    return series


def _coerce_boolean(series, null_invalid_values=False):
    if not pd.api.types.is_bool_dtype(series):
        series = _coerce_string(series).str.lower()
        return _transform_boolean(series, null_invalid_values)
    return series


def _transform_boolean(series, null_invalid_values):
    boolean_inference_list = config.get_option("boolean_inference_strings").copy()
    boolean_inference_list.update({frozenset(["1", "0"]), frozenset(["1.0", "0.0"])})
    boolean_transform_mappings = config.get_option("boolean_transform_mappings").copy()
    boolean_transform_mappings.update(
        {
            "1": True,
            "0": False,
            "1.0": True,
            "0.0": False,
        },
    )
    if null_invalid_values:
        series = series.apply(lambda x: boolean_transform_mappings.get(x, np.nan))
    else:
        series = series.apply(lambda x: boolean_transform_mappings.get(x, x))
    return series


def _coerce_integer(series):
    series = _coerce_numeric(series)
    invalid = _get_index_invalid_integer(series)
    if invalid.any():
        series[invalid] = None
    return series


def _coerce_age(series, fractional=False):
    coerce_type = _coerce_numeric if fractional else _coerce_integer
    series = coerce_type(series)
    invalid = _get_index_invalid_age(series)
    if invalid.any():
        series[invalid] = None
    return series


def _coerce_latlong(series):
    invalid = _get_index_invalid_latlong(series)
    if invalid.any():
        series[invalid] = None
    return series


def _coerce_postal_code(series):
    if pd.api.types.is_numeric_dtype(series):
        series = _coerce_integer(series).astype("Int64")
    return _coerce_string(series, regex="postal_code_inference_regex")



from pandas.core.dtypes.common import is_integer_dtype


def _get_numeric_value_counts_in_range(series, _range):
    """Count the number of occurrences of integers present in a series with values defined
    by a range of integers. Null values will be ignored.

    Args:
        series (pd.Series): data from which to determine the number of occurrences of values
        _range (type(range)): sequence of integers defining the values for which counts should be made

    Returns:
        value_counts (list(dict)): a list of dictionaries with keys `value` and
            `count`. Output is sorted in descending order based on the value counts.
    """
    frequencies = series.value_counts(dropna=True)
    value_counts = [
        {
            "value": i if is_integer_dtype(series) else float(i),
            "count": frequencies[i] if i in frequencies else 0,
        }
        for i in _range
    ]
    return sorted(value_counts, key=lambda i: (-i["count"], i["value"]))



def _convert_ordinal_to_numeric(schema, data):
    """Transforms ordinal data to numeric data so that dependence can be calculated.

    Args:
        schema (woodwork.TableSchema): Woodwork typing info for the data.
        data (dict[pd.Series]): Dictionary of Pandas series to use for
            calculating dependence.

    Returns:
        None
    """
    columns = [k for k, v in schema.logical_types.items() if "Ordinal" in str(v)]
    for col_name in columns:
        if col_name in data:
            order = list(schema.columns[col_name].logical_type.order)
            data[col_name] = data[col_name].apply(lambda x: order.index(x))



from woodwork.serializers.serializer_base import (
    PYARROW_IMPORT_ERROR_MESSAGE,
    Serializer,
    clean_latlong,
)
from woodwork.utils import import_or_raise


class OrcSerializer(Serializer):
    """Serialize a Woodwork table to disk as an orc file."""

    format = "orc"

    def serialize(self, dataframe, profile_name, **kwargs):
        import_or_raise("pyarrow", PYARROW_IMPORT_ERROR_MESSAGE)
        self.kwargs["engine"] = "pyarrow"
        return super().serialize(dataframe, profile_name, **kwargs)

    def write_dataframe(self):
        file = self._get_filename()
        dataframe = clean_latlong(self.dataframe)
        save_orc_file(dataframe, file)


def save_orc_file(dataframe, filepath):
    """Utility function to write dataframe to disk as orc file."""
    from pyarrow import Table, orc

    df = dataframe.copy()
    for c in df:
        if df[c].dtype.name == "category":
            df[c] = df[c].astype("string")
    pa_table = Table.from_pandas(df, preserve_index=False)
    orc.write_table(pa_table, filepath)



import json
import os
import tarfile
import tempfile
from inspect import getfullargspec
from pathlib import Path

from woodwork.deserializers import (
    ArrowDeserializer,
    CSVDeserializer,
    FeatherDeserializer,
    OrcDeserializer,
    ParquetDeserializer,
    PickleDeserializer,
)
from woodwork.s3_utils import get_transport_params, use_smartopen
from woodwork.utils import _is_s3, _is_url

FORMAT_TO_DESERIALIZER = {
    CSVDeserializer.format: CSVDeserializer,
    PickleDeserializer.format: PickleDeserializer,
    ParquetDeserializer.format: ParquetDeserializer,
    ArrowDeserializer.format: ArrowDeserializer,
    FeatherDeserializer.format: FeatherDeserializer,
    OrcDeserializer.format: OrcDeserializer,
}


def _get_deserializer(
    path,
    filename,
    data_subdirectory,
    typing_info_filename,
    profile_name,
    format,
):
    """Determine the proper Deserializer class to use based on the specified parameters.
    Initializes and returns the proper Deserializer object.

    Args:
        path (str): Directory on disk, S3 path, or URL to read data and typing information.
        filename (str, optional): The name of the file used to store the data during serialization. If not specified, will be
            determined from the typing info file. Must be specified when deserializing from a single parquet file.
        data_subdirectory (str, optional): The subdirectory in which the data was stored during serialization. Defaults to "data".
        typing_info_filename (str, optional): The name of the JSON file used to store the Woodwork typing information during
            serialization. Defaults to "woodwork_typing_info.json".
        format (str, optional): The format used to serialize the data. Required if the serialized filename suffix does not
            match the format.
        profile_name (str, bool): The AWS profile specified to write to S3. Will default to None and search for AWS credentials.
            Set to False to use an anonymous profile.

    Returns:
        Deserializer: Initialized `woodwork.Deserializer` object that can be used to deserialize data.
    """
    typing_info = None
    if typing_info_filename:
        try:
            typing_info = read_table_typing_information(
                path,
                typing_info_filename,
                profile_name,
            )
        except FileNotFoundError:
            pass

    if typing_info:
        format = typing_info["loading_info"]["type"]
    elif format is None and filename is not None:
        # Try to get format from filename suffix
        format = Path(filename).suffix[1:]

    deserializer_cls = FORMAT_TO_DESERIALIZER.get(format)
    if deserializer_cls is None:
        raise ValueError(
            "Could not determine format. Please specify filename and/or format.",
        )

    return deserializer_cls(path, filename, data_subdirectory, typing_info)


def read_table_typing_information(path, typing_info_filename, profile_name):
    """Read Woodwork typing information from disk, S3 path, or URL.

    Args:
        path (str): Location on disk, S3 path, or URL to read typing info file.
        typing_info_filename (str): Name of JSON file in which typing info is stored.
        profile_name (str, bool): The AWS profile specified to access to S3.

    Returns:
        dict: Woodwork typing information dictionary
    """
    if _is_url(path) or _is_s3(path):
        with tempfile.TemporaryDirectory() as tmpdir:
            file_name = Path(path).name
            file_path = os.path.join(tmpdir, file_name)
            transport_params = None

            if _is_s3(path):
                transport_params = get_transport_params(profile_name)

            use_smartopen(file_path, path, transport_params)
            with tarfile.open(str(file_path)) as tar:
                if "filter" in getfullargspec(tar.extractall).kwonlyargs:
                    tar.extractall(path=tmpdir, filter="data")
                else:
                    raise RuntimeError(
                        "Please upgrade your Python version to the latest patch release to allow for safe extraction of the EntitySet archive.",
                    )

            file = os.path.join(tmpdir, typing_info_filename)
            with open(file, "r") as file:
                typing_info = json.load(file)
    else:
        path = os.path.abspath(path)
        assert os.path.exists(path), '"{}" does not exist'.format(path)
        file = os.path.join(path, typing_info_filename)
        with open(file, "r") as file:
            typing_info = json.load(file)
        typing_info["path"] = path

    return typing_info
